import numpy as np
from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor


class GraphConstructorKNN(GraphConstructor):
    _estimator_type = "transformer"

    def __init__(self,
                 k_distance: int = 10,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for generating adjacency matrices
        from connectivity matrices. Selects the k nearest
        neighbours for each node based on pairwise distance.
        Recommended for functional connectivity.
        Adapted from Ktena et al, 2017.


        Parameters
        ----------
        k_distance: int
            the k nearest neighbours value, for the kNN algorithm.
        one_hot_nodes: int,default=0
            Whether to generate a one hot encoding of the nodes in the matrix (1) or not (0)
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int,default=0
            whether to perform a fisher transform of each matrix (1) or not (0)
        use_abs_fisher: int,default=0
            changes the values to absolute values. Is applied after fisher transform and before z-score transformation
        zscore: int,default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
        use_abs_zscore: int,default=0
            whether to use the absolute values of the z-score transformation or allow for negative values
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data

        Example
        -------
        Use outside of a PHOTON pipeline

        ```python
        constructor = GraphConstructorKNN(k_distance=6,
                                          fisher_transform=1,
                                          use_abs=1)
        ```

        Or as part of a pipeline

        ```python
        my_pipe.add(PipelineElement('GraphConstructorKNN',
                                    hyperparameters={'k_distance': 6}))
        ```
       """
        super(GraphConstructorKNN, self).__init__(one_hot_nodes=one_hot_nodes,
                                                  use_abs=use_abs,
                                                  fisher_transform=fisher_transform,
                                                  use_abs_fisher=use_abs_fisher,
                                                  zscore=zscore,
                                                  use_abs_zscore=use_abs_zscore,
                                                  adjacency_axis=adjacency_axis,
                                                  logs=logs)
        self.k_distance = k_distance

    def get_knn(self, adjacency: np.ndarray) -> np.ndarray:
        """Returns kNN matrices"""
        adjacency = np.squeeze(adjacency)
        adjacency_list = []
        for i in range(adjacency.shape[0]):
            # generate adjacency matrix
            d, idx = self.distance_sklearn_metrics(adjacency[i, :, :], k=self.k_distance, metric='euclidean')
            k_adjacency = self.adjacency(d, idx).astype(np.float32)

            # turn adjacency into numpy matrix for concatenation
            k_adjacency = k_adjacency.toarray()
            adjacency_list.append(k_adjacency)

        adjacency_knn = np.asarray(adjacency_list)
        adjacency_knn = adjacency_knn[:, :, :, np.newaxis]

        return adjacency_knn

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform matrices based on k nearest neighbours"""
        adj, feat = self.get_mtrx(X)
        # do preparatory matrix transformations
        adj = self.prep_mtrx(adj)
        # threshold matrix
        adj = self.get_knn(adj)
        # get feature matrix
        transformed = self.get_features(adj, feat)

        return transformed



from sklearn.base import BaseEstimator, TransformerMixin

from grakel import VertexHistogram, EdgeHistogram, ShortestPath, GraphletSampling, RandomWalk, NeighborhoodHash,\
    WeisfeilerLehman, NeighborhoodSubgraphPairwiseDistance, LovaszTheta, SvmTheta, OddSth, GraphHopper, Propagation,\
    PyramidMatch, SubgraphMatching, MultiscaleLaplacian, CoreFramework


class GrakelTransformer(BaseEstimator, TransformerMixin):

    transformations = {
        "VertexHistogram": VertexHistogram,
        "EdgeHistogram": EdgeHistogram,
        "ShortestPath": ShortestPath,
        "GraphletSampling": GraphletSampling,
        "RandomWalk": RandomWalk,
        "NeighborhoodHash": NeighborhoodHash,
        "WeisfeilerLehman": WeisfeilerLehman,
        "NeighborhoodSubgraphPairwiseDistance": NeighborhoodSubgraphPairwiseDistance,
        "LovaszTheta": LovaszTheta,
        "SvmTheta": SvmTheta,
        "OddSth": OddSth,
        "GraphHopper": GraphHopper,
        "Propagation": Propagation,
        "PyramidMatch": PyramidMatch,
        "SubgraphMatching": SubgraphMatching,
        "MultiscaleLaplacian": MultiscaleLaplacian,
        "CoreFramework": CoreFramework
    }

    def __new__(cls, transformation: str = None, **kwargs):
        if transformation not in cls.transformations:
            raise ValueError(f"The requested transformation {transformation} was not found.")
        return cls.transformations[transformation](**kwargs)



"""
===========================================================
Project: PHOTON Graph
===========================================================
Description
-----------
A wrapper containing functions for extracting photonai_graph measures that can then be
used for further machine learning analyses

Version
-------
Created:        09-09-2019
Last updated:   02-06-2020


Author
------
Vincent Holstein
Translationale Psychiatrie
Universitaetsklinikum Muenster
"""

import igraph
from tqdm.contrib.concurrent import thread_map
from functools import partial
import numpy as np
import json
import os

from photonai_graph.Measures.AbstractMeasureTransform import AbstractMeasureTransform


class IgraphMeasureTransform(AbstractMeasureTransform):
    _estimator_type = "transformer"

    def __init__(self,
                 graph_functions: dict = None,
                 adjacency_axis: int = 0,
                 logs: str = None,
                 n_processes: int = 1):
        """The GraphMeasureTransform class is a class for extracting graph measures from graphs.
        A range of networkx graph measures is available and can be used in a PHOTON pipeline or extracted and
        written to a csv file for further use.

        Parameters
        ----------
        graph_functions: dict,default=None
            a dict of graph functions to calculate with keys as the networkx function name and a dict of the arguments
            as the value. In this second dictionary, the keys are the functions arguments and values are the desired
            values for the argument.
        adjacency_axis: int,default=0
            Channel index for adjacency
        logs: str,default=None
            path to the log data
        n_processes: str,default=None
            Number of processes for multi processing

        Examples
        --------
        ```python
        measuretrans = GraphMeasureTransform(graph_functions={"large_clique_size": {},
                                                              "global_efficiency": {},
                                                              "overall_reciprocity": {},
                                                              "local_efficiency": {}})
        ```
        """
        super(IgraphMeasureTransform, self).__init__(graph_functions=graph_functions)
        if self.graph_functions is None:
            self.graph_functions = {"degree": {}, "eigenvector_centrality": {}}
        self.n_processes = n_processes
        self.adjacency_axis = adjacency_axis

        self.logs = logs
        if not logs:
            self.logs = os.getcwd()

    def fit(self, X, y):
        return self

    def _inner_transform(self, X):
        x_transformed = []

        graphs = X

        # load json file
        base_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        measure_json = os.path.join(base_folder, 'IgraphMeasures.json')
        with open(measure_json, 'r') as measure_json_file:
            measure_j = json.load(measure_json_file)

        if self.n_processes > 1:
            pfn = partial(self._compute_graph_metrics, graph_functions=self.graph_functions, measure_j=measure_j)
            x_transformed = thread_map(pfn, graphs, max_workers=self.n_processes)
        else:
            for graph in graphs:
                measure_list_graph = self._compute_graph_metrics(graph, self.graph_functions, measure_j)
                x_transformed.append(measure_list_graph)

        return self._shared_inner_transform(x_transformed=x_transformed)

    def transform(self, X):
        if not isinstance(X, (list, np.ndarray, np.matrix)):
            raise TypeError('Input needs to list/ndarray of Igraph objects')
        if not isinstance(X[0], igraph.Graph):
            raise TypeError('Input needs to be list/ndarray or Igraph objects')

        x_transformed = self._inner_transform(X)
        return self._shared_transform(x_transformed=x_transformed)

    def _compute_graph_metrics(self, graph, graph_functions, measure_j):
        measure_list_graph = []
        for key, value in graph_functions.items():
            measure_list = list()

            if key not in measure_j:
                raise ValueError(f"Measure function {key} not found")

            measure = measure_j[key]

            # call function
            results = getattr(graph, measure["path"].split(".")[-1])(**value)
            measure_list = self.handle_outputs(results, measure_list)

            if "compute_average" in measure.keys() and measure['compute_average']:
                measure_list_graph.append([np.mean(measure_list)])
            else:
                measure_list_graph.append(measure_list)
        return measure_list_graph

    def extract_measures(self, x_graphs_in, path="", ids=None):
        x_graphs = x_graphs_in.copy()
        if ids is None:
            raise ValueError('No id provided')
        self._shared_extraction(x_graphs, ids, path)



import unittest
import numpy as np
import igraph
from photonai_graph.GraphConversions import dense_to_igraph


class DenseToIgraphTest(unittest.TestCase):

    def setUp(self):
        self.random4d = np.random.rand(20, 20, 20, 2)

    def test_axis_argument(self):
        with self.assertRaises(NotImplementedError):
            dense_to_igraph(self.random4d, feature_axis=1)

    def test_matrix(self):
        g = dense_to_igraph(self.random4d, adjacency_axis=0)
        self.assertEqual(type(g[0]), igraph.Graph)
        self.assertEqual(type(g), list)

    def test_node_count(self):
        g = dense_to_igraph(self.random4d, adjacency_axis=0)
        self.assertEqual(g[0].vcount(), 20)



from sklearn.base import BaseEstimator, TransformerMixin
from typing import Tuple, Union
import numpy as np


class PopulationAveragingTransform(BaseEstimator, TransformerMixin):
    def __init__(self,
                 adjacency_axis: int = 0,
                 connectivity_axis: int = 1,
                 feature_axis: Union[int, Tuple[int, ...]] = 2):
        """Transformation for averaging over the population
        The connectivity of an individual is discarded by this transformation.
        Instead, the precalculated mean is used and the original connectivity will be passed as features.

        !!! warning
            The original adjacency matrix of unseen data is discarded by application of this transformer!

        Parameters
        ----------
        adjacency_axis: int,default=0
            Index of the adjacency part of the input arrays.
            If a GraphConstructor is used with default parameters the adjacency is stored at the first dimension.
        connectivity_axis: int,default=1
            Index of the connectivity part of the input arrays for unseen data.
            If a GraphConstructor is used with default parameters the connectivity is at the second dimension.
        feature_axis: Union[int, List[int, ...]],default=2
            Index (or indices) of the feature part of the unseen data.

        Example
        -------
        ```python
        my_pipe.add(PipelineElement('PopulationAveragingTransform'))
        ```

        Notes
        -----
        Using the feature axis as additional feature is not implemented yet.
        """
        self.connectivtiy_axis = connectivity_axis
        self.adjacency_axis = adjacency_axis
        self.feature_axis = feature_axis
        self.learned_mean = None
        if self.feature_axis != 2:  # pragma: no cover
            raise NotImplementedError("Feature passing is not implemented yet.")

    def fit(self, X: np.ndarray, y: np.ndarray):
        if np.ndim(X) != 4:
            raise ValueError("Unexpected input dimensionality")
        mean = np.mean(X, axis=0)
        self.learned_mean = mean[..., self.adjacency_axis]

        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        if self.learned_mean is None:
            raise ValueError("This transformer has not been fitted yet.")
        if np.ndim(X) != 4:
            raise ValueError("Unexpected input dimensionality.")
        new_adjacency = self.learned_mean.copy()
        new_adjacency = new_adjacency[np.newaxis, ..., np.newaxis]
        new_adjacency = np.repeat(new_adjacency, X.shape[0], axis=0)
        return np.concatenate((new_adjacency, X[..., self.feature_axis, np.newaxis]), axis=3)



import unittest
import numpy as np
from photonai_graph.GraphEmbedding.graph_embedding_locally_linear_embedding import GraphEmbeddingLocallyLinearEmbedding


class LocallyLinearEmbeddingTest(unittest.TestCase):

    def setUp(self):
        mtrx = np.random.rand(20, 20, 20, 2)
        mtrx[:, :, :, 0][mtrx[:, :, :, 0] <= 0.9] = 0
        self.X = mtrx
        self.y = np.ones(20)

    def test_embedding_locally_linear_embedding(self):
        g_embedding = GraphEmbeddingLocallyLinearEmbedding(embedding_dimension=1)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertEqual((20, 20), np.shape(gembed))

    def test_embedding_laplacian_eigenmaps_complex(self):
        g_embedding = GraphEmbeddingLocallyLinearEmbedding(embedding_dimension=1)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        bool_mask = [False] * 20
        bool_array = np.iscomplex(gembed[0])
        comp = bool_mask == bool_array
        comp = comp.all()
        self.assertTrue(comp)

    def test_embedding_locally_linear_embedding_3d(self):
        g_embedding = GraphEmbeddingLocallyLinearEmbedding(embedding_dimension=3)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertEqual((20, 20, 3), np.shape(gembed))



import warnings
from abc import ABC, abstractmethod
import os
from photonai_graph.util import assert_imported
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.model_selection import train_test_split
from tqdm import tqdm
import numpy as np
try:
    import dgl
    import torch
    import torch.nn as nn
    import torch.optim as optim
    from torch.utils.data import DataLoader
    from dgl.dataloading import GraphDataLoader
except ImportError:
    import warnings
    warnings.warn("Some of the dependencies could not be loaded. Make sure to install all Dependencies:\n"
                  "https://wwu-mmll.github.io/photonai_graph/installation/#additional-packages")

from photonai_graph.GraphConversions import check_dgl
from photonai_graph.NeuralNets.NNUtilities import DGLData, zip_data


class DGLModel(BaseEstimator, ABC):
    _estimator_type = "classifier"

    def __init__(self, nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 early_stopping: bool = False,
                 es_patience: int = 10,
                 es_tolerance: int = 9,
                 es_delta: float = 0,
                 verbose: bool = False,
                 logs: str = None):
        """
        Base class for DGL based graph neural networks. Implements
        helper functions and shared parameters used by other models.
        Implementation based on dgl python package.


        Parameters
        ----------
        nn_epochs: int,default=200
            the number of epochs which a model is trained
        learning_rate: float,default=0.001
            the learning rate when training the model
        batch_size: int,default=32
            number of samples per training batch
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero
        feature_axis: int,default=1
            position of the feature matrix
        add_self_loops: bool,default=True
            self loops are added if true
        allow_zero_in_degree: bool,default=False
            If true the zero in degree test of dgl is disabled
        validation_score: bool,default=False
            If true the input data is split into train and test (90%/10%).
            The testset is then used to get validation results during training
        early_stopping: bool, default=False
            If true then the loss over multiple iterations is evaluated to see
            whether early stopping should be called on the model
        verbose: bool,default=False
            If true verbose information is printed
        logs: str,default=None
            Path to the log data

        """
        self.nn_epochs = nn_epochs
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.adjacency_axis = adjacency_axis
        self.feature_axis = feature_axis
        self.add_self_loops = add_self_loops
        self.allow_zero_in_degree = allow_zero_in_degree
        self.validation_score = validation_score
        self.early_stopping = early_stopping
        self.es_patience = es_patience
        self.es_tolerance = es_tolerance
        self.es_delta = es_delta
        if self.add_self_loops and self.allow_zero_in_degree:
            warnings.warn('If self loops are added allow_zero_in_degree should be false!')
        if not self.add_self_loops and not self.allow_zero_in_degree:
            warnings.warn('If no self loops are added and allow_zero_in_degree is set to false, '
                          'all graphs should not contain 0-in-degree nodes')
        self.model = None
        if logs:
            self.logs = logs
        else:
            self.logs = os.getcwd()
        assert_imported(["dgl", "pytorch"])
        self.verbose = verbose

    def train_model(self, epochs, model, optimizer, loss_func, data_loader, val_loader=None):
        # This function trains the neural network
        epoch_losses = []
        val_losses = []
        for epoch in tqdm(range(epochs)):
            epoch_loss = 0
            iteration = 0
            for it, (bg, label) in enumerate(data_loader):
                prediction = model(bg)
                loss = loss_func(prediction, label)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                epoch_loss += loss.detach().item()
                iteration = it
            epoch_loss /= (iteration + 1)
            if self.verbose and not self.validation_score:
                print('Epoch {} \tloss {:.4f}'.format(epoch, epoch_loss))
            if self.verbose and self.validation_score:
                val_loss = 0
                for it, (bg, label) in enumerate(val_loader):
                    prediction = model(bg)
                    loss = loss_func(prediction, label)
                    val_loss += loss.detach().item()
                    iteration = it
                val_loss /= (iteration + 1)
                print(f'Epoch {epoch} \tloss {epoch_loss:.4f} \tval loss {val_loss:.4f}')
                val_losses.append(val_loss)
            epoch_losses.append(epoch_loss)
            convergence = self.check_val_loss_divergence(val_losses, epoch_losses,
                                                         self.es_patience, self.es_tolerance, self.es_delta)
            if not convergence:
                break

    def fit(self, X, y=None):
        # handle inputs
        transformed_inputs = self.handle_inputs(X, self.adjacency_axis, self.feature_axis)
        # get data loader
        val_loader = None
        if not self.validation_score:
            data_loader = self.get_data_loader(transformed_inputs, y)
        else:
            print("10% of your train data is used as validation data!\ndisable with validation_score=False")
            x_train, x_val, y_train, y_val = train_test_split(transformed_inputs, y, test_size=.1)
            data_loader = self.get_data_loader(x_train, y_train)
            val_loader = self.get_data_loader(x_val, y_val)
        # specify model with optimizer etc
        self._init_model(X, y)
        # get optimizers
        loss_func, optimizer = self.setup_model()
        # train model
        self.model.train()
        self.train_model(epochs=self.nn_epochs, model=self.model, optimizer=optimizer,
                         loss_func=loss_func, data_loader=data_loader, val_loader=val_loader)
        return self

    def handle_inputs(self, x, adjacency_axis, feature_axis):
        """checks the format of the input and transforms them for dgl models"""
        x_trans = check_dgl(x, adjacency_axis=adjacency_axis, feature_axis=feature_axis)
        if self.add_self_loops:
            x_trans = [dgl.add_self_loop(x) for x in x_trans]
        return x_trans

    def predict(self, x):
        return self.predict_model(x)

    def get_data_loader(self, x_trans, y):
        data = DGLData(zip_data(x_trans, y))
        data_loader = DataLoader(data, batch_size=self.batch_size, shuffle=True, collate_fn=self.collate)
        return data_loader

    @staticmethod
    def check_val_loss_divergence(val_loss_list: list, ep_loss_list: list,
                                  patience: int = 10, tolerance: int = 9,
                                  delta=0):
        if len(val_loss_list) < 10:
            convergence = True
        else:
            val_recent = val_loss_list[-patience:]
            ep_recent = ep_loss_list[-patience:]
            train_val_divergence = [val - ep for val, ep in zip(val_recent, ep_recent)]
            counter = sum(i > delta for i in train_val_divergence)
            if counter >= tolerance:
                convergence = False
            else:
                convergence = True

        return convergence

    @staticmethod
    @abstractmethod
    def collate(samples):
        """Collate function"""

    @abstractmethod
    def predict_model(self, X):
        """get model predictions"""

    @abstractmethod
    def setup_model(self):
        """Setup the model"""

    @abstractmethod
    def _init_model(self, X=None, y=None):
        """initialize model"""


class DGLClassifierBaseModel(DGLModel, ClassifierMixin, ABC):
    def __init__(self,
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 verbose: bool = False,
                 logs: str = None,
                 **kwargs):
        """Abstract base class for classification algorithms

                Parameters
                ----------
                nn_epochs: int,default=200
                    Number of epochs to fit the model
                learning_rate: float,default=0.001
                    Learning rate for model training
                batch_size: int,default=32
                    Batch size for model training
                adjacency_axis: int,default=0
                    Axis which contains the adjacency
                feature_axis: int,default=1
                    Axis which contains the features
                add_self_loops: bool,default=True
                    If this value is true, a self loop is added to each node of each graph
                allow_zero_in_degree: bool,default=False
                    If true the dgl model allows zero-in-degree Graphs
                validation_score: bool,default=False
                    It true the input data is split into train and test (90%/10%).
                    The testset is then used to get validation results during training
                verbose: bool,default=False
                    If true verbose output is generated
                logs: str,default=None
                    Default logging directory
                """
        super(DGLClassifierBaseModel, self).__init__(nn_epochs=nn_epochs,
                                                     learning_rate=learning_rate,
                                                     batch_size=batch_size,
                                                     adjacency_axis=adjacency_axis,
                                                     feature_axis=feature_axis,
                                                     add_self_loops=add_self_loops,
                                                     allow_zero_in_degree=allow_zero_in_degree,
                                                     validation_score=validation_score,
                                                     verbose=verbose,
                                                     logs=logs,
                                                     **kwargs)

    def setup_model(self):
        """returns the loss and optimizer for classification"""
        loss_func = nn.CrossEntropyLoss()  # specify loss
        optimizer = optim.Adam(self.model.parameters(), lr=self.learning_rate)
        return loss_func, optimizer

    def predict_model(self, X):
        """returns the argmax of the predictions for classification tasks"""
        self.model.eval()
        x_trans = self.handle_inputs(X, self.adjacency_axis, self.feature_axis)
        test_bg = dgl.batch(x_trans)
        probs_y = torch.softmax(self.model(test_bg), 1)
        argmax_y = torch.max(probs_y, 1)[1].view(-1, 1)
        return argmax_y.squeeze()

    @staticmethod
    def collate(samples):
        """returns a batched graph, the input (samples) is a list of pairs (graph, label)"""
        graphs, labels = map(list, zip(*samples))
        batched_graph = dgl.batch(graphs)
        return batched_graph, torch.tensor(labels, dtype=torch.long)


class DGLRegressorBaseModel(DGLModel, RegressorMixin, ABC):
    def __init__(self,
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 verbose: bool = False,
                 logs: str = None,
                 **kwargs):
        """Abstract base class for regression algorithms

        Parameters
        ----------
        nn_epochs: int,default=200
            Number of epochs to fit the model
        learning_rate: float,default=0.001
            Learning rate for model training
        batch_size: int,default=32
            Batch size for model training
        adjacency_axis: int,default=0
            Axis which contains the adjacency
        feature_axis: int,default=1
            Axis which contains the features
        add_self_loops: bool,default=True
            If this value is true, a self loop is added to each node of each graph
        allow_zero_in_degree: bool,default=False
            If true the dgl model allows zero-in-degree Graphs
        validation_score: bool,default=False
            If true the input data is split into train and test (90%/10%).
            The testset is then used to get validation results during training
        verbose: bool,default=False
            If true verbose output is generated
        logs: str,default=None
            Default logging directory
        """
        super(DGLRegressorBaseModel, self).__init__(nn_epochs=nn_epochs,
                                                    learning_rate=learning_rate,
                                                    batch_size=batch_size,
                                                    adjacency_axis=adjacency_axis,
                                                    feature_axis=feature_axis,
                                                    add_self_loops=add_self_loops,
                                                    allow_zero_in_degree=allow_zero_in_degree,
                                                    validation_score=validation_score,
                                                    verbose=verbose,
                                                    logs=logs,
                                                    **kwargs)

    def setup_model(self):
        """returns the loss and optimizer for regression"""
        loss_func = nn.MSELoss()  # regression loss
        optimizer = optim.Adam(self.model.parameters(), lr=self.learning_rate)
        return loss_func, optimizer

    def predict_model(self, X):
        self.model.eval()
        x_trans = self.handle_inputs(X, self.adjacency_axis, self.feature_axis)
        test_bg = dgl.batch(x_trans)
        probs = self.model(test_bg)
        probs = probs.detach().numpy()
        return probs.squeeze()

    def get_data_loader(self, x_trans, y):
        """returns data in a regression data loader format"""
        y = y.reshape(y.shape[0], 1)
        return super(DGLRegressorBaseModel, self).get_data_loader(x_trans=x_trans, y=y)

    @staticmethod
    def collate(samples):
        """returns a batched graph, the input (samples) is a list of pairs"""
        graphs, labels = map(list, zip(*samples))
        batched_graph = dgl.batch(graphs)
        return batched_graph, torch.tensor(labels, dtype=torch.float32)



import unittest
import os
import numpy as np
import pandas as pd

from photonai_graph.Controllability.controllability_measures import ControllabilityMeasureTransform


class ControllabilityTransformTests(unittest.TestCase):

    def setUp(self):
        b = np.random.randint(2, size=(20, 20))
        b_symm = (b + b.T) / 2
        b_symm[b_symm < 0.6] = 0
        b_symm = b_symm[np.newaxis, :, :, np.newaxis]
        b_symm = np.repeat(b_symm, 10, axis=0)
        self.X_sym = b_symm
        self.y = np.random.rand(20)
        self.rep_X_sym = np.load(os.path.dirname(__file__) + '/X_test.npz')['arr_0']
        self.path = "./test.csv"

    def test_mod_control_shape(self):
        contr = ControllabilityMeasureTransform(mod_control=1, ave_control=0)
        contr.fit(self.X_sym, self.y)
        vec = contr.transform(self.X_sym)
        self.assertEqual(np.shape(vec), (10, 20))

    def test_ave_control_shape(self):
        # ensure that mean transform style with a 3d matrix returns the right shape
        contr = ControllabilityMeasureTransform(mod_control=0, ave_control=1)
        contr.fit(self.X_sym, self.y)
        vec = contr.transform(self.X_sym)
        self.assertEqual(np.shape(vec), (10, 20))

    def test_mod_ave_control_shape(self):
        contr = ControllabilityMeasureTransform(mod_control=1, ave_control=1)
        contr.fit(self.X_sym, self.y)
        vec = contr.transform(self.X_sym)
        self.assertEqual(np.shape(vec), (10, 40))

    def test_error_transform(self):
        # ensure that mean transform style with a 3d matrix returns the right shape
        with self.assertRaises(ValueError):
            ControllabilityMeasureTransform(mod_control=0, ave_control=0)

    def test_extract_measures_mod_control(self):
        contr = ControllabilityMeasureTransform(mod_control=1, ave_control=0)
        contr.extract_measures(self.X_sym, self.path)
        os.remove(self.path)

    def test_extract_measures_ave_control(self):
        contr = ControllabilityMeasureTransform(mod_control=0, ave_control=1)
        contr.extract_measures(self.X_sym, self.path)
        os.remove(self.path)

    def test_extract_measures_mod_ave_control(self):
        contr = ControllabilityMeasureTransform(mod_control=1, ave_control=1)
        contr.extract_measures(self.rep_X_sym, self.path, ids=[i for i in range(10)],
                               node_list=[f"node_{i}" for i in range(20)])
        df = pd.read_csv(self.path)
        path_ref = os.path.dirname(__file__) + "/test.csv"
        df_expected = pd.read_csv(path_ref)
        self.assertTrue(np.allclose(df.to_numpy(), df_expected.to_numpy()), "Generated measures are not as expected")
        os.remove(self.path)

    def test_node_list_error(self):
        node_list = list(range(5))
        contr = ControllabilityMeasureTransform(mod_control=1, ave_control=0)
        with self.assertRaises(ValueError):
            contr.extract_measures(self.X_sym, self.path, node_list=node_list)

    def test_id_list_error(self):
        id_list = list(range(3))
        contr = ControllabilityMeasureTransform(mod_control=1, ave_control=0)
        with self.assertRaises(ValueError):
            contr.extract_measures(self.X_sym, self.path, ids=id_list)



from sklearn.base import BaseEstimator, TransformerMixin
import networkx as nx
import igraph
import numpy as np
import os

from photonai_graph.GraphConversions import dense_to_networkx, dense_to_igraph
from photonai_graph.util import NetworkxGraphWrapper


class GraphMeasureAdapter(BaseEstimator, TransformerMixin):
    _estimator_type = "transformer"

    def __init__(self,
                 output: str = None,
                 adjacency_axis: int = 0,
                 logs: str = None,
                 n_processes: int = 1):
        """The GraphMeasureTransform class is a class for extracting graph measures from graphs.
        A range of networkx graph measures is available and can be used in a PHOTON pipeline or extracted and
        written to a csv file for further use.

        Parameters
        ----------
        graph_functions: dict,default=None
            a dict of graph functions to calculate with keys as the networkx function name and a dict of the arguments
            as the value. In this second dictionary, the keys are the functions arguments and values are the desired
            values for the argument.
        adjacency_axis: int,default=0
            Channel index for adjacency
        logs: str,default=None
            path to the log data
        n_processes: str,default=None
            Number of processes for multi processing

        Examples
        --------
        ```python
        measuretrans = GraphMeasureTransform(graph_functions={"large_clique_size": {},
                                                              "global_efficiency": {},
                                                              "overall_reciprocity": {},
                                                              "local_efficiency": {}})
        ```
        """
        self.n_processes = n_processes
        self.output = output
        self.adjacency_axis = adjacency_axis

        self.logs = logs
        if not logs:
            self.logs = os.getcwd()

    def fit(self, X, y):
        return self

    def _inner_transform(self, X):

        if self.output == 'igraph':
            if isinstance(X, np.ndarray) or isinstance(X, np.matrix):
                graphs = dense_to_igraph(X, adjacency_axis=self.adjacency_axis)
            elif isinstance(X, list) and min([isinstance(g, igraph.Graph) for g in X]):
                graphs = X
            else:
                raise TypeError("Input needs to a list of igraph graphs or numpy array")

        elif self.output == 'networkx':
            if isinstance(X, np.ndarray) or isinstance(X, np.matrix):
                graphs = dense_to_networkx(X, adjacency_axis=self.adjacency_axis)
            elif isinstance(X, list) and min([isinstance(g, nx.Graph) for g in X]):
                graphs = X
            else:
                raise TypeError("Input has to be a list of networkx graphs or numpy array.")

            graphs = [NetworkxGraphWrapper(g) for g in graphs]

        else:
            raise NotImplementedError('Can only convert to networkx or igraph.'
                                      'Please ensure spelling is correct.')

        return graphs

    def transform(self, X):
        transformed = self._inner_transform(X)
        transformed = np.asarray(transformed)
        return transformed



import unittest
import numpy as np
from photonai_graph.GraphConstruction.graph_constructor_threshold import GraphConstructorThreshold


class ThresholdTests(unittest.TestCase):

    def setUp(self):
        self.X4d_adjacency = np.ones((20, 20, 20, 1))
        self.X4d_features = np.random.rand(20, 20, 20, 1)
        self.X4d = np.concatenate((self.X4d_adjacency, self.X4d_features), axis=3)
        self.test_mtrx = np.array(([0.5, 0, 0, 0],
                                  [0, 1, 0, 0],
                                  [0, 0, 1, 0],
                                  [0, 0, 0, 0.5]))
        self.y = np.ones(20)

    def test_wrong_input_shape(self):
        g_constr = GraphConstructorThreshold(threshold=.5)
        input_mtrx = np.ones((15, 20, 20))
        g_constr.fit(input_mtrx, np.arange(15))
        with self.assertRaises(ValueError):
            g_constr.transform(input_mtrx)

    def test_strange_one_hot_value(self):
        with self.assertRaises(ValueError):
            GraphConstructorThreshold(one_hot_nodes=27.3)

    def test_threshold_4d(self):
        # ensure that individual transform style with a 4d matrix returns the right shape
        g_constr = GraphConstructorThreshold(threshold=0.5)
        g_constr.fit(self.X4d, self.y)
        trans = g_constr.transform(self.X4d)
        self.assertEqual(trans.shape, (20, 20, 20, 3))
        # first dimension should be thresholded but unchanged
        self.assertTrue(np.array_equal(trans[..., 0, np.newaxis], self.X4d_adjacency))
        # second dimension should contain original connectivity
        self.assertTrue(np.array_equal(trans[..., 1, np.newaxis], self.X4d_adjacency))
        # last dimension should contain the random features
        self.assertTrue(np.array_equal(trans[..., 2, np.newaxis], self.X4d_features))

    def test_threshold_4d_discard_connectivity(self):
        # ensure that individual transform style with a 4d matrix returns the right shape
        g_constr = GraphConstructorThreshold(threshold=0.5,
                                             discard_original_connectivity=True)
        g_constr.fit(self.X4d, self.y)
        trans = g_constr.transform(self.X4d)
        self.assertEqual(trans.shape, (20, 20, 20, 2))
        # first dimension should be thresholded but unchanged
        self.assertTrue(np.array_equal(trans[..., 0, np.newaxis], self.X4d_adjacency))
        # last dimension should contain the random features
        self.assertTrue(np.array_equal(trans[..., 1, np.newaxis], self.X4d_features))

    def test_threshold_shape_4d_onehot(self):
        # ensure that an individual transform with a 3d matrix returns the right shape
        # when using one hot encoded features
        g_constr = GraphConstructorThreshold(threshold=0.5, one_hot_nodes=1)
        g_constr.fit(self.X4d, self.y)
        trans = g_constr.transform(self.X4d)
        self.assertEqual(trans.shape, (20, 20, 20, 4))
        # the first dimension still contains the (thresholded but unchanged) values
        self.assertTrue(np.array_equal(trans[..., 0, np.newaxis], self.X4d_adjacency))
        # the second dimesion contains the one hot encoding
        # We know the one hot encoding, as we created the matrix accordingly
        self.assertTrue(np.array_equal(trans[..., 1, np.newaxis],
                                       np.repeat(np.eye(20)[np.newaxis, ...], 20, axis=0)[..., np.newaxis]))
        # the third dimension contains again the original values
        self.assertTrue(np.array_equal(trans[..., 2, np.newaxis], self.X4d_adjacency))
        # the last dimension contains the features
        self.assertTrue(np.array_equal(trans[..., 3, np.newaxis], self.X4d_features))

    def test_threshold_individual_shape_4d_onehot_discard_connectivity(self):
        # ensure that an individual transform with a 3d matrix returns the right shape
        # when using one hot encoded features
        g_constr = GraphConstructorThreshold(threshold=0.5,
                                             one_hot_nodes=1,
                                             discard_original_connectivity=True)
        g_constr.fit(self.X4d, self.y)
        trans = g_constr.transform(self.X4d)
        self.assertEqual(trans.shape, (20, 20, 20, 3))
        # the first dimension still contains the (thresholded but unchanged) values
        self.assertTrue(np.array_equal(trans[..., 0, np.newaxis], self.X4d_adjacency))
        # the second dimesion contains the one hot encoding
        # We know the one hot encoding, as we created the matrix accordingly
        self.assertTrue(np.array_equal(trans[..., 1, np.newaxis],
                                       np.repeat(np.eye(20)[np.newaxis, ...], 20, axis=0)[..., np.newaxis]))
        # the last dimension contains the features
        self.assertTrue(np.array_equal(trans[..., 2, np.newaxis], self.X4d_features))

    def test_prep_matrix(self):
        g_constr = GraphConstructorThreshold(threshold=.0, use_abs=True)
        input_matrix = np.eye(4)
        output_matrix = g_constr.prep_mtrx(input_matrix * -1)
        self.assertTrue(np.array_equal(input_matrix, output_matrix))

    def test_use_abs(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, use_abs=True)
        input_matrix = np.eye(4) * -1
        output_matrix = g_constr.prep_mtrx(input_matrix)
        self.assertTrue(np.array_equal(np.eye(4), output_matrix))

    def test_use_abs_fisher(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, fisher_transform=1, use_abs_fisher=1)
        input_matrix = np.eye(4) * -1
        input_matrix = input_matrix[np.newaxis, :, :, np.newaxis]
        output_matrix = g_constr.prep_mtrx(input_matrix)
        ids = np.diag(output_matrix[0, ..., 0])
        self.assertTrue(np.array_equal(np.isposinf(ids), [True, True, True, True]))

    def test_use_zscore(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, zscore=1)
        output_matrix = g_constr.prep_mtrx(self.test_mtrx[np.newaxis, :, :, np.newaxis])
        self.assertEqual((np.sum(np.array(output_matrix) >= 0)), 4)

    def test_use_abs_zscore(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, zscore=1, use_abs_zscore=1)
        output_matrix = g_constr.prep_mtrx(self.test_mtrx[np.newaxis, :, :, np.newaxis])
        self.assertEqual((np.sum(np.array(output_matrix) >= 0)), 16)

    def test_use_abs_and_fisher(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, use_abs=1, fisher_transform=1)
        output_matrix = g_constr.prep_mtrx(self.test_mtrx[np.newaxis, :, :, np.newaxis])
        self.assertEqual((np.sum(np.array(output_matrix) >= 1)), 2)

    def test_use_abs_and_fisher_and_abs_fisher(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, use_abs=1, fisher_transform=1, use_abs_fisher=1)
        output_matrix = g_constr.prep_mtrx(self.test_mtrx[np.newaxis, :, :, np.newaxis])
        self.assertEqual((np.sum(np.array(output_matrix) >= 1)), 2)

    def test_use_abs_and_fisher_and_zsore(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, use_abs=1, fisher_transform=1, zscore=1)
        output_matrix = g_constr.prep_mtrx(self.test_mtrx[np.newaxis, :, :, np.newaxis])
        self.assertEqual((np.sum(np.array(output_matrix) >= 1)), 2)

    def test_fisher_and_zscore_and_abszscore(self):
        g_constr = GraphConstructorThreshold(threshold=0.5, fisher_transform=1, zscore=1, use_abs_zscore=1)
        output_matrix = g_constr.prep_mtrx(self.test_mtrx[np.newaxis, :, :, np.newaxis])
        self.assertEqual((np.sum(np.array(output_matrix) >= 1)), 2)



import unittest
import numpy as np
import networkx as nx
import grakel
from photonai_graph.GraphKernels.grakel_adapter import GrakelAdapter


class GrakelAdapterTests(unittest.TestCase):

    def setUp(self):
        mtrx = np.random.rand(20, 20, 20, 2)
        mtrx[:, :, :, 0][mtrx[:, :, :, 0] <= 0.9] = 0
        self.Xrandom4d = mtrx
        self.Xrandom_missfit = np.random.rand(20, 20, 5, 2)
        self.X_nx = [nx.erdos_renyi_graph(20, p=0.3)] * 20
        self.y = np.ones(20)

    def test_grakel_adapter_type_list(self):
        g_adapter = GrakelAdapter()
        g_adapter.fit(self.Xrandom4d, self.y)
        x_trans = g_adapter.transform(self.Xrandom4d)
        self.assertEqual(type(x_trans), list)

    def test_unexpected_input(self):
        g_adapter = GrakelAdapter(feature_axis=0)
        g_in = np.array([[1, 0, 0], [-1, 3, 4]])
        with self.assertRaises(ValueError):
            g_adapter.transform(g_in[np.newaxis, ..., np.newaxis])

    def test_unexpected_input_type(self):
        with self.assertRaises(ValueError):
            GrakelAdapter(input_type="non-sens")

    def test_grakel_adapter_type_graph(self):
        g_adapter = GrakelAdapter()
        x_trans = g_adapter.transform(self.Xrandom4d)
        self.assertEqual(type(x_trans[0]), grakel.Graph)

    def test_grakel_adapter_integration_kernel_shape(self):
        g_adapter = GrakelAdapter()
        g_kernel = grakel.kernels.PyramidMatch()
        x_trans = g_adapter.transform(self.Xrandom4d)
        x_kern = g_kernel.fit_transform(x_trans)
        self.assertEqual(x_kern.shape, (20, 20))

    def test_grakel_adapter_adjacency_shape(self):
        with self.assertRaises(Exception):
            g_adapter = GrakelAdapter()
            g_adapter.transform(self.Xrandom_missfit)

    def test_grakel_adapter_networkx(self):
        g_adapter = GrakelAdapter(input_type="networkx")
        g_kernel = grakel.kernels.RandomWalk()
        x_trans = g_adapter.transform(self.X_nx)
        x_kern = g_kernel.fit_transform(x_trans)
        self.assertEqual(x_kern.shape, (20, 20))



import numpy as np
try:
    from gem.embedding.hope import HOPE
    from gem.embedding.static_graph_embedding import StaticGraphEmbedding
except ImportError:  # pragma: no cover
    pass

from photonai_graph.GraphEmbedding.graph_embedding_base import GraphEmbeddingBase


class GraphEmbeddingHOPE(GraphEmbeddingBase):
    _estimator_type = "transformer"

    def __init__(self,
                 embedding_dimension: int = 1,
                 decay_factor: float = 0.01,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for calculating a Graph Embedding
        based on Higher-order proximity preserved embedding
        (Mingdong et al., 2016).
        Implementation based on gem python package.


        Parameters
        ----------
        embedding_dimension: int,default=1
            the number of dimensions that the final embedding will have
        decay_factor: float,default=0.01
            the higher order coefficient beta
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero
        logs: str,default=None
            Path to the log data



        Example
        -------
            constructor = GraphEmbeddingHOPE(embedding_dimension=1,
                                             decay_factor=0.1)
        """
        super(GraphEmbeddingHOPE, self).__init__(embedding_dimension=embedding_dimension,
                                                 adjacency_axis=adjacency_axis,
                                                 logs=logs)
        self.decay_factor = decay_factor
        self.orig_transformed = None

    def transform(self, X: np.ndarray) -> np.ndarray:
        x_transformed = super(GraphEmbeddingHOPE, self).transform(X)
        self.orig_transformed = x_transformed
        if self.embedding_dimension == 1:
            x_transformed = np.squeeze(np.reshape(x_transformed, (X.shape[0], -1, 1)))
        return x_transformed

    def _init_embedding(self) -> StaticGraphEmbedding:
        embedding_dimension = self.embedding_dimension
        if embedding_dimension == 1:
            embedding_dimension = 2
        return HOPE(d=embedding_dimension, beta=self.decay_factor)



import unittest

import numpy as np
import networkx as nx
from photonai_graph.PopulationAveragingTransform import PopulationAveragingTransform


class PopulationAveragingTests(unittest.TestCase):

    def setUp(self):
        # Randomly create 10 barabasi albert graphs and store the adjacency as numpy array
        self.graphs = np.array([nx.adjacency_matrix(nx.barabasi_albert_graph(20, 2)).todense() for _ in range(10)])
        # Generate random graphs for 10 subjects containing 3 measurements
        self.other_graphs = np.random.rand(10, 20, 20, 3)

    def test_averaging(self):
        mean = np.squeeze(np.mean(self.graphs, axis=0))
        pat = PopulationAveragingTransform()
        pat.fit(self.graphs[..., np.newaxis], 0)
        self.assertTrue(np.array_equal(mean, pat.learned_mean))

        transformed = pat.transform(self.other_graphs)
        for subject in range(transformed.shape[0]):
            self.assertTrue(np.array_equal(mean, transformed[subject, ..., 0]))
            self.assertTrue(np.array_equal(self.other_graphs[subject, ..., 2], transformed[subject, ..., 1]))

    def test_unfitted_transformer(self):
        pat = PopulationAveragingTransform()
        with self.assertRaises(ValueError):
            pat.transform(self.other_graphs)

    def test_wrong_input_dimensionality(self):
        dummy_input = np.ones((10, 20, 20))
        pat = PopulationAveragingTransform()
        with self.assertRaises(ValueError):
            pat.fit(dummy_input, [])
        pat.fit(dummy_input[..., np.newaxis], [])
        with self.assertRaises(ValueError):
            pat.transform(dummy_input)



import unittest
import numpy as np
import networkx as nx
from dgl.data import MiniGCDataset
from photonai_graph.NeuralNets.GCNModel import GCNClassifierModel
from photonai_graph.GraphUtilities import get_random_labels


class GCNClassifierTests(unittest.TestCase):

    def setUp(self):
        mtrx = np.random.rand(20, 20, 20, 2)
        mtrx[:, :, :, 0][mtrx[:, :, :, 0] <= 0.8] = 0
        for i in range(mtrx.shape[0]):
            mtrx[i, :, :, 0][np.diag_indices_from(mtrx[i, :, :, 0])] = 1
        self.Xrandom4d = mtrx
        self.Xrandom_missfit = np.random.rand(20, 20, 5, 2)
        self.X_nx = [nx.erdos_renyi_graph(20, p=0.3)] * 20
        self.y = get_random_labels(number_of_labels=20)
        dgl_mini = MiniGCDataset(20, 10, 20)
        self.X_dgl = dgl_mini.graphs
        self.y_mini = dgl_mini.labels

    def test_gcn_classifier_output_shape(self):
        gcn_clf = GCNClassifierModel(nn_epochs=20)
        gcn_clf.fit(self.Xrandom4d, self.y)
        output = gcn_clf.predict(self.Xrandom4d)
        self.assertTrue(np.array_equal(output.shape, self.y.shape))

    def test_gcn_classifier_output_valset(self):
        gcn_clf = GCNClassifierModel(nn_epochs=20, verbose=True, validation_score=True)
        gcn_clf.fit(self.Xrandom4d, self.y)
        output = gcn_clf.predict(self.Xrandom4d)
        self.assertTrue(np.array_equal(output.shape, self.y.shape))

    def test_gcn_classifier_output_hidden_dim(self):
        gcn_clf = GCNClassifierModel(hidden_dim=128)
        gcn_clf.fit(self.Xrandom4d, self.y)
        output = gcn_clf.predict(self.Xrandom4d)
        self.assertTrue(np.array_equal(output.shape, self.y.shape))

    def test_gat_classifier_dgl(self):
        gat_clf = GCNClassifierModel(nn_epochs=20)
        gat_clf.fit(self.X_dgl, self.y)
        output = gat_clf.predict(self.X_dgl)
        self.assertTrue(np.array_equal(np.array(output.shape), self.y.shape))



import numpy as np

from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor


class GraphConstructorThreshold(GraphConstructor):
    _estimator_type = "transformer"

    def __init__(self,
                 threshold: float = 0.1,
                 concatenation_axis: int = 3,
                 return_adjacency_only: int = 0,
                 retain_weights: int = 0,
                 discard_original_connectivity: bool = False,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for generating adjacency matrices
        from connectivity matrices. Thresholds matrix based
        on a chosen threshold value.


        Parameters
        ----------
        threshold: float
            threshold value below which to set matrix entries to zero
        concatenation_axis: int
            axis along which to concatenate the adjacency and feature matrix
        return_adjacency_only: int
            whether to return the adjacency matrix only (1) or also a feature matrix (0)
        retain_weights: int
            whether to retain weight values or not
        discard_original_connectivity: bool,default=False
            If true the original connectivity will be passed on. Otherwise it gets discarded
        one_hot_nodes: int
            Whether to generate a one hot encoding of the nodes in the matrix (1) or not (0)
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int
            whether to perform a fisher transform of each matrix (1) or not (0)
        use_abs_fisher: int, default=0
            changes the values to absolute values. Is applied after fisher transform and before z-score transformation
        zscore: int, default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
        use_abs_zscore:
            whether to use the absolute values of the z-score transformation or allow for negative values
        adjacency_axis: int, default=0
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data


        Example
        -------
        Use outside of a PHOTON pipeline

        ```python
        constructor = GraphConstructorThreshold(threshold=0.5,
                                                fisher_transform=1,
                                                use_abs=1)
        ```

        Or as part of a pipeline

        ```python
        my_pipe.add(PipelineElement('GraphConstructorThreshold',
                                    hyperparameters={'threshold': 0.5}))
        ```
       """
        super(GraphConstructorThreshold, self).__init__(one_hot_nodes=one_hot_nodes,
                                                        use_abs=use_abs,
                                                        fisher_transform=fisher_transform,
                                                        discard_original_connectivity=discard_original_connectivity,
                                                        use_abs_fisher=use_abs_fisher,
                                                        zscore=zscore,
                                                        use_abs_zscore=use_abs_zscore,
                                                        adjacency_axis=adjacency_axis,
                                                        logs=logs)
        if retain_weights not in [0, 1]:
            raise ValueError("retain_weights has to be in [0, 1]")
        self.threshold = threshold
        self.concatenation_axis = concatenation_axis
        self.return_adjacency_only = return_adjacency_only
        self.retain_weights = retain_weights

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform input matrices using a threshold"""
        adj, feat = self.get_mtrx(X)
        # do preparatory matrix transformations
        adj = self.prep_mtrx(adj)
        # threshold matrix
        adj = self.threshold_matrix(adj)
        # get feature matrix
        transformed = self.get_features(adj, feat)

        return transformed

    def threshold_matrix(self, adjacency: np.ndarray) -> np.ndarray:
        """Threshold matrix"""
        adjacency[adjacency < self.threshold] = 0
        if not self.retain_weights:
            adjacency[adjacency >= self.threshold] = 1

        return adjacency



from scipy.linalg import svd, schur
import numpy as np


def modal_control(a_mat):
    # This method returns values of MODAL CONTROLLABILITY for each node
    # in a network, given the adjacency matrix for that network. Modal
    # controllability indicates the ability of that node to steer the
    # system into difficult-to-reach states, given input at that node.
    #
    # INPUT:
    #     a_mat is the structural (NOT FUNCTIONAL) network adjacency matrix,
    # 	such that the simple linear model of dynamics outlined in the
    # 	reference is an accurate estimate of brain state fluctuations.
    # 	Assumes all values in the matrix are positive, and that the
    # 	matrix is symmetric.
    #
    # OUTPUT:
    #     Vector of modal controllability values for each node
    #
    # Reference: Gu, Pasqualetti, Cieslak, Telesford, Yu, Kahn, Medaglia,
    #            Vettel, Miller, Grafton & Bassett, Nature Communications
    #            6:8414, 2015.
    u, s, vt = svd(a_mat)  # singluar value decomposition
    a_mat = a_mat / (1 + s[0])  # s is a eigen-value
    t_mat, u_mat = schur(a_mat, 'real')  # Schur stability
    eig_vals = np.diag(t_mat)
    n_mat = a_mat.shape[0]
    phi = np.zeros(n_mat, dtype=float)
    for i in range(n_mat):
        a_left = u_mat[i, ] * u_mat[i, ]  # element-wise multiplication
        a_right = (1.0 - np.power(eig_vals, 2)).transpose()
        phi[i] = np.matmul(a_left, a_right)
    return phi


def average_control(a_mat):
    # This program is a Python version of average controllability
    #     This function returns values of AVERAGE CONTROLLABILITY for
    #     each node in a network, given the adjacency matrix for that network.
    #     Average controllability measures the ease by which input at
    #     that node can steer the system into many easily-reachable states.
    #
    # INPUT:
    #     A is the structural (NOT FUNCTIONAL) network adjacency matrix,
    #     such that the simple linear model of dynamics outlined in the%
    #     reference is an accurate estimate of brain state fluctuations.
    #     Assumes all values in the matrix are positive, and that the
    #     matrix is symmetric.
    #
    # OUTPUT:
    #     Vector of average controllability values for each node
    #
    #     Reference: Gu, Pasqualetti, Cieslak, Telesford, Yu, Kahn, Medaglia,
    #             Vettel, Miller, Grafton & Bassett, Nature Communications
    #             6:8414, 2015.
    # %%
    n_mat = a_mat.shape[0]
    u, s, vt = svd(a_mat)  # singluar value decomposition
    a_mat = a_mat / (1 + s[0])  # s is a eigen-value
    t_mat, u_mat = schur(a_mat, 'real')  # Schur stability
    mid_mat = np.multiply(u_mat, u_mat).transpose()  # element-wise multiplication
    v = np.matrix(np.diag(t_mat)).transpose()

    # %%
    p_mat = np.diag(1 - np.matmul(v, v.transpose()))
    p_mat = np.tile(p_mat.reshape([a_mat.shape[0], 1]), (1, n_mat))
    values = sum(np.divide(mid_mat, p_mat))
    return values



import unittest
import os
import numpy as np
import pandas as pd
from photonai_graph.GraphConstruction.graph_constructor_spatial import GraphConstructorSpatial


class SpatialTests(unittest.TestCase):

    def setUp(self):
        # get random matrix as feature matrix
        self.X4d = np.ones((20, 12, 12, 2))
        self.Xrandom4d = np.random.rand(20, 12, 12, 2)
        self.XrandomHO4d = np.random.rand(20, 110, 110, 2)
        self.y = np.ones((20))

        # get test atlas
        coord_dict = {'x': [39, 42, 42, 45, 3, 6, 6, 9, 42, 39, 45, 42],
                      'y': [45, 39, 42, 42, 3, 3, 3, 3, 3, 3, 3, 3],
                      'z': [45, 45, 39, 42, 42, 39, 45, 48, 3, 6, 6, 9]}
        self.spatial_coords = pd.DataFrame(data=coord_dict)

        # get check array
        arr = np.array([[3]])
        self.check_array = np.tile(arr, (1, 12))

        # get check array for ho_atlas
        self.ho_check = [4, 3, 4, 5, 5, 4, 3, 4, 3, 4, 5,
                         4, 4, 5, 3, 3, 5, 5, 3, 4, 3, 3,
                         4, 4, 4, 4, 3, 3, 3, 3, 3, 4, 4,
                         3, 4, 4, 4, 3, 4, 3, 3, 3, 3, 3,
                         3, 3, 3, 3, 3, 3, 4, 4, 3, 3, 4,
                         4, 4, 4, 3, 3, 5, 5, 3, 4, 4, 3,
                         6, 6, 4, 4, 4, 4, 3, 3, 5, 6, 6,
                         7, 3, 4, 3, 3, 3, 3, 3, 4, 3, 3,
                         4, 3, 4, 3, 4, 3, 4, 4, 4, 5, 4,
                         5, 5, 4, 4, 4, 4, 4, 6, 4, 3, 3]

    def test_spatial(self):
        path = "./test_coords.csv"
        self.spatial_coords.to_csv(path, header=False, index=False)
        g_constr = GraphConstructorSpatial(k_distance=3, atlas_name='test', atlas_folder="./")
        g_constr.fit(self.Xrandom4d, self.y)
        trans = g_constr.transform(self.Xrandom4d)
        self.assertEqual(np.shape(trans), (20, 12, 12, 3))
        os.remove(path)

    def test_spatial_mechanism(self):
        path = "./test_coords.csv"
        self.spatial_coords.to_csv(path, header=False, index=False)
        g_constr = GraphConstructorSpatial(k_distance=3, atlas_name='test', atlas_folder="./")
        g_constr.fit(self.Xrandom4d, self.y)
        trans = g_constr.transform(self.Xrandom4d)
        bool_array = np.count_nonzero(trans[0, :, :, 0], axis=1, keepdims=True) == self.check_array
        self.assertFalse(False in bool_array)
        os.remove(path)

    def test_ho_atlas(self):
        path = os.path.dirname(os.path.abspath(__file__))
        g_constr = GraphConstructorSpatial(k_distance=3, atlas_name='ho', atlas_folder=path)
        g_constr.fit(self.XrandomHO4d, self.y)
        trans = g_constr.transform(self.XrandomHO4d)
        ho_adj = trans[0, :, :, 0]
        self.assertTrue(np.array_equal(np.count_nonzero(ho_adj, axis=0), self.ho_check))
            
    def test_ho_atlas_sum(self):
        path = os.path.dirname(os.path.abspath(__file__))
        g_constr = GraphConstructorSpatial(k_distance=3, atlas_name='ho', atlas_folder=path)
        g_constr.fit(self.XrandomHO4d, self.y)
        trans = g_constr.transform(self.XrandomHO4d)
        ho_adj = trans[0, :, :, 0]
        self.assertTrue((np.sum(ho_adj) > 190))

    def test_NotImplementedError(self):
        g_constr = GraphConstructorSpatial(k_distance=3, atlas_name='test', atlas_folder=None)
        with self.assertRaises(NotImplementedError):
            g_constr.transform(self.XrandomHO4d)



import os
from typing import Union, List

from networkx.drawing.nx_pylab import draw
import networkx.drawing as drawx
from networkx.algorithms import asteroidal
import networkx as nx
import numpy as np
import pydot
from scipy import stats
from scipy import sparse
import matplotlib.pyplot as plt
import scipy.io as sio
try:
    import mat73
except ImportError:  # pragma: no cover
    pass

from photonai_graph.util import assert_imported


def draw_connectogram(graph, connection_style="arc3", colorscheme=None, nodesize=None,
                      node_shape='o', weight=None, path=None, show=True):
    """This functions draws a connectogram, from a graph.

    Parameters
    ----------
    graph: nx.class.graph.Graph
        input graph, a single networkx graph
    connection_style:
        connection style, controlling the style of the drawn edges
    colorscheme:
        colormap for drawing the connectogram
    nodesize: int
        controls size of the drawn nodes
    node_shape: str, default='o'
        shape of the drawn nodes
    weight: float
        threshold below which edges are coloured differently than above
    path: str, default=None
        path where to save the plots as string, if no path is declared, plots are not saved.
        Path needs to be the full path including file name and ending, unlike in draw_connectograms
    show: bool, default=True
        whether to plot the graph or not. Set it to false in headless environments
    """

    pos = nx.circular_layout(graph)
    if colorscheme is not None:
        nx.draw_networkx_nodes(graph, pos, node_size=nodesize, node_color=range(nx.number_of_nodes(graph)),
                               node_shape=node_shape, cmap=plt.get_cmap(colorscheme))
    else:
        nx.draw_networkx_nodes(graph, pos, node_size=nodesize, node_shape=node_shape, cmap=plt.get_cmap(colorscheme))

    if weight is not None:
        elarge = [(u, v) for (u, v, d) in graph.edges(data=True) if d['weight'] > weight]
        esmall = [(u, v) for (u, v, d) in graph.edges(data=True) if d['weight'] <= weight]
        nx.draw_networkx_edges(graph, pos, edgelist=elarge,
                               connectionstyle=connection_style)
        nx.draw_networkx_edges(graph, pos, edgelist=esmall,
                               alpha=0.5, edge_color='b', style='dashed', connectionstyle=connection_style)

    else:
        nx.draw_networkx_edges(graph, pos, connectionstyle=connection_style)

    if show:  # pragma: no cover
        plt.show()

    if path is not None:
        plt.savefig(path)


def draw_connectograms(graphs, curved_edge=False, colorscheme=None,
                       nodesize=None, node_shape='o', weight=None,
                       path=None, ids=None, out_format=None, show=True):
    """This function draws multiple connectograms, from graph lists.

    Parameters
    ----------
    graphs
        input graphs, a list of networkx graphs or a single networkx graph
    curved_edge: bool, default=False
        whether to draw straight or curved edges
    colorscheme:
        colormap for drawing the connectogram
    nodesize: int
        controls size of the drawn nodes
    node_shape: str, default='o'
        shape of the drawn nodes
    weight: float
        threshold below which edges are coloured differently than above
    path: str, default=None
        path where to save the plots as string, if no path is declared, plots are not saved
    ids: list, default=None
        list of ids, after which to name the plots
    out_format: str, default=None
        output format for the graphs, as a string
    show: bool, default=True
        whether to plot the connectograms or not. Set it to false in headless environments
    """
    if isinstance(graphs, list):
        if ids is not None:
            if len(ids) == len(graphs):
                for graph, current_id in zip(graphs, ids):
                    if None in [path, out_format]:
                        raise ValueError('To save graphs, declare a path and an output format.')
                    save_path = os.path.join(path, str(current_id) + out_format)
                    draw_connectogram(graph, curved_edge, colorscheme, path=save_path, show=show)
            else:
                raise ValueError('Number of IDs must match number of graphs.')
        # if no IDs are provided graphs are just numbered
        else:
            counter = 0
            if None in [path, out_format]:
                for graph in graphs:
                    draw_connectogram(graph, curved_edge, colorscheme, nodesize, node_shape, weight, show=show)
                    counter += 1
            else:
                for graph in graphs:
                    save_path = os.path.join(path, str(counter) + out_format)
                    draw_connectogram(graph, curved_edge, colorscheme, nodesize, node_shape, weight, path=save_path, show=show)
                    counter += 1

    # if the it is only a single graph
    elif isinstance(graphs, nx.classes.graph.Graph):
        draw_connectogram(graphs, curved_edge, colorscheme, path=path, show=show)

    # input should be list or single graph
    else:
        raise TypeError('Input needs to be a single networkx graph or a list of those.')


def draw_connectivity_matrix(matrix, colorbar=False, colorscheme="viridis", adjacency_axis=None, show=True):
    """Draw connectivity matrix.

        Parameters
        ----------
        matrix : numpy.ndarray, numpy.matrix or a list of those
        the input matrix or matrices from which to draw the connectivity matrix

        colorbar : boolean, default=False
            Whether to use a colorbar in the drawn plot

        colorscheme: str, default="viridis"
            colorscheme for plotting the connectivity matrix

        adjacency_axis : int, default=None
        position of the the adjacency axis, if specified the array is assumed to
        have an additional axis where the matrix is stored.

        show: bool, default=True
            whether to show the connectivity matrix or not.

        Notes
        -----
        If new nodes are added with features, and any of the old nodes
        do not have some of the feature fields, those fields are filled
        by initializers defined with ``set_n_initializer`` (default filling
        with zeros).

        Examples
        --------
        >>> g = get_random_connectivity_data()
        >>> draw_connectivity_matrix(adjacency_axis=0)

        """
    # check input format
    if isinstance(matrix, np.ndarray) or isinstance(matrix, np.matrix):
        if adjacency_axis is not None:
            if np.ndim(matrix) == 4:
                for i in range(matrix.shape[0]):
                    plt.imshow(matrix[i, :, :, adjacency_axis], cmap=plt.get_cmap(colorscheme))
                    if colorbar:
                        plt.colorbar()
                    if show:  # pragma: no cover
                        plt.show()
            elif np.ndim(matrix) == 3:
                plt.imshow(matrix[:, :, adjacency_axis], cmap=plt.get_cmap(colorscheme))
                if colorbar:
                    plt.colorbar()
                if show:  # pragma: no cover
                    plt.show()
            else:
                raise ValueError('Matrix dimension might not be specified correctly.')
        else:
            if np.ndim(matrix) == 4:
                raise ValueError('You have 4 dimension, please specify axis to plot')
            elif np.ndim(matrix) == 3:
                for i in range(matrix.shape[0]):
                    plt.imshow(matrix[i, :, :])
                    if colorbar:
                        plt.colorbar()
                    if show:  # pragma: no cover
                        plt.show()
            elif np.ndim(matrix) == 2:
                plt.imshow(matrix)
                if colorbar:
                    plt.colorbar()
                if show:  # pragma: no cover
                    plt.show()
            else:
                raise ValueError('Matrix dimension might not be specified correctly.')
    elif isinstance(matrix, list):
        if isinstance(matrix[0], np.ndarray) or isinstance(matrix[0], np.matrix):
            for single_matrix in matrix:
                if adjacency_axis is not None:
                    plt.imshow(single_matrix[:, :, adjacency_axis])
                    if colorbar:
                        plt.colorbar()
                    if show:  # pragma: no cover
                        plt.show()
        elif isinstance(matrix[0], sparse.spmatrix) \
                or isinstance(matrix[0], sparse.bsr_matrix) \
                or isinstance(matrix[0], sparse.lil_matrix) \
                or isinstance(matrix[0], sparse.csc_matrix) \
                or isinstance(matrix[0], sparse.coo_matrix) \
                or isinstance(matrix[0], sparse.csr_matrix) \
                or isinstance(matrix[0], sparse.dok_matrix) \
                or isinstance(matrix[0], sparse.dia_matrix):
            for single_matrix in matrix:
                plt.spy(single_matrix)
                if show:  # pragma: no cover
                    plt.show()
        else:
            raise TypeError('List elements need to be numpy arrays/matrices or scipy sparse matrices')
    else:
        raise TypeError('draw_connectivity_matrix only takes numpy arrays, matrices or lists as input.')


def get_random_connectivity_data(out_type="dense",
                                 number_of_nodes=114,
                                 number_of_individuals=10,
                                 number_of_modalities=2):
    """generate random connectivity matrices for testing and debugging

        Parameters
        ----------
        out_type: str, default="dense"
            output type for connectivity data, default="dense"
        number_of_nodes: int, default=114
            number of nodes in the matrix/graph
        number_of_individuals: int, default=10
            number of individual graphs/matrices
        number_of_modalities: int, default=2
            number of modalities as per matrix/graph
    """
    if out_type == "dense":
        random_matrices = np.random.rand(number_of_individuals, number_of_nodes, number_of_nodes, number_of_modalities)
    elif out_type == "sparse":
        random_matrices = []
        for _ in range(number_of_individuals):
            modality_list = []
            for _ in range(number_of_modalities):
                random_matrix = sparse.random(number_of_nodes, number_of_nodes, density=0.1)
                modality_list.append(random_matrix)
            random_matrices.append(modality_list)
    else:
        raise NotImplementedError("Only dense and sparse matrices are supported as output type.")

    return random_matrices


def get_random_labels(l_type="classification", number_of_labels=10):
    """get random labels for testing and debugging functions.

        Parameters
        ----------
        l_type : str, default="classification"
            controls the type labels. "classification" outputs binary labels 0 and 1, "regression" outputs random float.

        number_of_labels : int, default=10
            number of labels to generate

        Returns
        -------
        np.ndarray
            The labels as np.ndarray


        Notes
        -----
        If used in conjunction with get_random_connectivity_data number_of_labels
        should match number_of_individuals.

        Examples
        --------
        >>> labels = get_random_labels()

        """

    if l_type == "classification" or l_type == "Classification":
        y = np.random.rand(number_of_labels)
        y[y > 0.5] = 1
        y[y < 0.5] = 0

    elif l_type == "regression" or l_type == "Regression":
        y = np.random.rand(number_of_labels)

    else:
        raise ValueError('random labels only implemented for classification and regression. Please check your spelling')

    return y


def visualize_networkx(graphs, layout=nx.spring_layout, colorscheme="Blues", show=True):
    """Visualize a networkx graph or graphs using networkx built-in visualization.

        Parameters
        ----------
        graphs :
            a list or of networkx graphs or a single networkx graph
        layout :
            layout of the graph, default is spring layout
        colorscheme: str, default="Blues"
            colormap for the nodes, default is Blues
        show: bool, default=True
            whether to show the plot or not. Set it to False in headless environments

        Examples
        --------
        >>> g1, g2 = nx.line_graph(10), nx.line_graph(7)
        >>> graphs = [g1, g2]
        >>> visualize_networkx(graphs)

        """
    # check format in which graphs are presented or ordered
    if isinstance(graphs, list):
        for graph in graphs:
            draw(graph, pos=layout(graph), node_color=range(nx.number_of_nodes(graph)), cmap=plt.get_cmap(colorscheme))
            if show:  # pragma: no cover
                plt.show()
    elif isinstance(graphs, nx.classes.graph.Graph):
        draw(graphs, pos=layout(graphs), node_color=range(nx.number_of_nodes(graphs)), cmap=plt.get_cmap(colorscheme))
        if show:  # pragma: no cover
            plt.show()
    else:
        raise ValueError("graphs has unexpected format")


def individual_ztransform(matrx, adjacency_axis=0):
    """applies a z-score transformation individually to each connectivity matrix in an array

        Parameters
        ----------
        matrx : np.ndarray
            a list or of networkx graphs or a single networkx graph

        adjacency_axis: int, default=0
            the position of the adjacency matrix

        Returns
        -------
        np.ndarray
            The z-score transformed matrices as an array


        Examples
        --------
        >>> X = get_random_connectivity_data()
        >>> X_transformed = individual_ztransform(X)

        """
    # check dimensions
    transformed_matrices = []
    if np.ndim(matrx) != 4:
        raise ValueError("Please check input dimension of your graph data.")
    for i in range(matrx.shape[0]):
        matrix = matrx[i, :, :, adjacency_axis].copy()
        matrix = stats.zscore(matrix)
        transformed_matrices.append(matrix)

    transformed_matrices = np.asarray(transformed_matrices)

    return transformed_matrices


def individual_fishertransform(matrx, adjacency_axis=0):
    """applies a fisher transformation individually to each connectivity matrix in an array.
        all values need to be finite (non nans) as this would introduce complex numbers, which
        can not be handled by other methods.

        Parameters
        ----------
        matrx : np.ndarray
            An array containing the adjacency values for the graphs
        adjacency_axis: int, default=0
            the position of the adjacency matrix

        Returns
        -------
        np.ndarray
            The fisher transformed matrices as an array


        Examples
        --------
        >>> X = get_random_connectivity_data()
        >>> X_transformed = individual_fishertransform(X)

        """
    # check dimensions
    transformed_matrices = []
    if np.ndim(matrx) != 4:
        raise ValueError("Please check input dimension of your graph data.")
    if not np.isfinite(matrx).all():
        raise ValueError("For a fisher transform graph data should only include finite values, no nans")
    for i in range(matrx.shape[0]):
        matrix = matrx[i, :, :, adjacency_axis].copy()
        matrix = np.arctanh(matrix)
        transformed_matrices.append(matrix)

    transformed_matrices = np.asarray(transformed_matrices)
    transformed_matrices = transformed_matrices[:, :, :, np.newaxis]

    return transformed_matrices


def pydot_to_nx(graphs):
    """transforms pydot graphs to networkx graphs

        Parameters
        ----------
        graphs : list or pydot.Dot
        a list of pydot graphs or a single pydot graph


        Returns
        -------
        list or pydot.Dot
            The input graph or graphs in networkx format

        """
    if isinstance(graphs, list):
        a_graphs = []
        for graph in graphs:
            a_graph = drawx.nx_pydot.from_pydot(graph)
            a_graphs.append(a_graph)

    elif isinstance(graphs, pydot.Dot):
        a_graphs = drawx.nx_pydot.from_pydot(graphs)

    else:
        raise TypeError('The input needs to be list of pydot files or a single pydot file. Please check your inputs.')

    return a_graphs


def check_asteroidal_boolean(graph: Union[nx.Graph, List[nx.Graph]]) -> Union[bool, List[bool]]:
    """checks wether a graph or a list of graphs is asteroidal

    Parameters
    ----------
    graph: Union[nx.classes.graph.Graph, List[nx.classes.graph.Graph]]
        Input graph(s)

    Returns
    -------
    Union[bool, List[bool]]
        True if graph is asteroidal, False otherwise per graph
    """
    if isinstance(graph, list):
        graph_answer = []
        for i in graph:
            answer = asteroidal.is_at_free(i)
            graph_answer.append(answer)
    elif isinstance(graph, nx.classes.graph.Graph):
        graph_answer = asteroidal.is_at_free(graph)
    else:
        raise ValueError('Your input is not a networkx photonai_graph or a list of networkx graphs. '
                         'Please check your inputs.')
    return graph_answer


def check_asteroidal(graph: Union[nx.Graph, List[nx.Graph]]) -> Union[Union[List, None], List[Union[List, None]]]:
    """checks whether a graph or a list of graphs is asteroidal

        Parameters
        ----------
        graph : list or nx.classes.graph.Graph
            a list of networkx graphs or a single networkx graph

        return_boolean : boolean, default=True
            whether to return a True/False statement about the graphs or a list of asteroidal triples per graph


        Returns
        -------
        list
            A list of asteroidal triples or a single asteroidal triple

        """
    # checks for asteroidal triples in the photonai_graph or in a list of networkx graphs
    if isinstance(graph, list):
        graph_answer = []
        for i in graph:
            answer = asteroidal.find_asteroidal_triple(i)
            graph_answer.append(answer)
    elif isinstance(graph, nx.classes.graph.Graph):
        graph_answer = asteroidal.find_asteroidal_triple(graph)
    else:
        raise ValueError('Your input is not a networkx photonai_graph or a list of networkx graphs. '
                         'Please check your inputs.')

    return graph_answer


def load_conn(path='', mtrx_name='matrix', subject_dim=3, modality_dim=2):  # pragma: no cover
    """loads matlab 4d connectivity matrix from matlab file
    This function is untested and serves only as an example how to load connectivity data.
    """
    assert_imported(['mat73'])
    try:
        matfile = sio.loadmat(path)
        mtrx = matfile[mtrx_name]
        mtrx = np.moveaxis(mtrx, [subject_dim, modality_dim], [0, 3])
    except Exception as ex1:
        try:
            matfile = mat73.loadmat(path)
            mtrx = matfile[mtrx_name]
            mtrx = np.moveaxis(mtrx, [subject_dim, modality_dim], [0, 3])
        except Exception as ex2:
            raise IOError(ex1, ex2)

    return mtrx



import unittest
from tempfile import mkstemp

import numpy as np
import networkx as nx
import os


from photonai_graph.Measures.NetworkxMeasureTransform import NetworkxMeasureTransform
from photonai_graph.util import NetworkxGraphWrapper


class GraphMeasureTransformTests(unittest.TestCase):

    def setUp(self):
        gs = np.load(os.path.dirname(__file__) + '/X_test.npz')['arr_0']
        self.X_nx = [NetworkxGraphWrapper(nx.from_numpy_array(gs[i])) for i in range(gs.shape[0])]
        self.y = np.random.rand(10)
        self.ids = list(range(10))
        # generate random matrices
        self.random_mtrx = np.random.rand(10, 20, 20, 2)
        # generate nonsense input
        self.edge_dict = {(1, 0): 1, (2, 0): 1, (3, 0): 1}

    def test_transform_nonsense_input(self):
        g_transform = NetworkxMeasureTransform()
        with self.assertRaises(TypeError):
            g_transform.fit(self.edge_dict, self.y)
            g_transform.transform(self.edge_dict)

    def test_transform_default_measures(self):
        g_transform = NetworkxMeasureTransform()
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 2))

    def test_transform_own_measures(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"global_efficiency": {},
                                                                "local_efficiency": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 2))

    def test_transform_own_measures_parallel(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"global_efficiency": {},
                                                                "local_efficiency": {}},
                                               n_processes=2)
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 2))

    def test_transform_dict_node(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"eigenvector_centrality": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 20))

    def test_transform_dict_edge(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"edge_current_flow_betweenness_centrality": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(type(measures), np.ndarray)

    def test_transform_list(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"voterank": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(type(measures), np.ndarray)

    def test_transform_dict_dict(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"communicability": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 400))

    def test_transform_float_or_dict(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"clustering": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 20))

    def test_transform_tuple_dict(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"hits": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 40))

    def test_transform_dual_tuple(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"non_randomness": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 2))

    def test_transform_directed(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"degree_pearson_correlation_coefficient": {}})
        g_transform.fit(self.X_nx, self.y)
        measures = g_transform.transform(self.X_nx)
        self.assertEqual(measures.shape, (10, 1))

    def test_extract_input_mtrx(self):
        _, path = mkstemp()
        g_transform = NetworkxMeasureTransform()
        g_transform.fit(self.random_mtrx, self.y)
        g_transform.extract_measures(self.random_mtrx, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_extract_nonsense_input_no_ids(self):
        g_transform = NetworkxMeasureTransform()
        with self.assertRaises(ValueError):
            g_transform.fit(self.edge_dict, self.y)
            g_transform.extract_measures(self.edge_dict)

    def test_extract_default_measures(self):
        _, path = mkstemp('.csv')
        g_transform = NetworkxMeasureTransform()
        g_transform.extract_measures(self.X_nx, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_extract_own_measures(self):
        _, path = mkstemp('.csv')
        g_transform = NetworkxMeasureTransform(graph_functions={"global_efficiency": {},
                                                                "local_efficiency": {}})
        g_transform.extract_measures(self.X_nx, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_extract_dict_node(self):
        _, path = mkstemp('.csv')
        g_transform = NetworkxMeasureTransform(graph_functions={"eigenvector_centrality": {}})
        g_transform.extract_measures(self.X_nx, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_extract_dict_edge(self):
        _, path = mkstemp('.csv')
        g_transform = NetworkxMeasureTransform(graph_functions={"edge_current_flow_betweenness_centrality": {}})
        g_transform.extract_measures(self.X_nx, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_extract_float_or_dict(self):
        _, path = mkstemp('.csv')
        g_transform = NetworkxMeasureTransform(graph_functions={"clustering": {}})
        g_transform.extract_measures(self.X_nx, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_extract_no_id(self):
        g_transform = NetworkxMeasureTransform()
        with self.assertRaises(Exception):
            g_transform.extract_measures(self.X_nx)

    def test_extract_directed(self):
        _, path = mkstemp('.csv')
        g_transform = NetworkxMeasureTransform(graph_functions={"degree_pearson_correlation_coefficient": {}})
        g_transform.fit(self.X_nx, self.y)
        g_transform.extract_measures(self.X_nx, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_compute_average(self):
        g_transform = NetworkxMeasureTransform(graph_functions={"degree_centrality": {}})
        g_transform.fit(self.X_nx, self.y)
        x_nodes = g_transform.transform(self.X_nx)
        x_nodes_average = np.mean(x_nodes, axis=1).reshape(-1, 1)

        g_transform = NetworkxMeasureTransform(graph_functions={"average_degree_centrality": {}})
        g_transform.fit(self.X_nx, self.y)
        x_average = g_transform.transform(self.X_nx)

        np.testing.assert_array_equal(x_average, x_nodes_average)

    # !!!
    #   Below this line are legacy tests where
    #   NotImplementedErrors are expected.
    #   ---------------------------------------
    #   todo: Reimplement all these tests!
    # !!!

    """def test_extract_list(self):
        path = "/tmp/test.csv"
        g_transform = GraphMeasureTransform(graph_functions={"voterank": {}})
        with self.assertRaises(NotImplementedError):
            g_transform.extract_measures(self.X_nx, path, self.ids)"""

    """def test_extract_dict_dict(self):
        path = "/tmp/test.csv"
        g_transform = GraphMeasureTransform(graph_functions={"communicability": {}})
        with self.assertRaises(NotImplementedError):
            g_transform.extract_measures(self.X_nx, path, self.ids)"""

    """def test_extract_tuple_dict(self):
        path = "/tmp/test.csv"
        g_transform = GraphMeasureTransform(graph_functions={"hits": {}})
        with self.assertRaises(NotImplementedError):
            g_transform.extract_measures(self.X_nx, path, self.ids)"""

    """    
    def test_extract_dual_tuple(self):
        path = "/tmp/test.csv"
        g_transform = GraphMeasureTransform(graph_functions={"non_randomness": {}})
        with self.assertRaises(NotImplementedError):
            g_transform.extract_measures(self.X_nx, path, self.ids)"""



import unittest
import numpy as np
from photonai_graph.GraphEmbedding.graph_embedding_hope import GraphEmbeddingHOPE


class HOPETest(unittest.TestCase):

    def setUp(self):
        mtrx = np.random.rand(20, 20, 20, 2)
        mtrx[:, :, :, 0][mtrx[:, :, :, 0] <= 0.9] = 0
        self.X = mtrx
        self.y = np.ones(20)

    def test_embedding_hope(self):
        g_embedding = GraphEmbeddingHOPE()
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertEqual((20, 40), np.shape(gembed))
        orig_transformed = g_embedding.orig_transformed
        for graph in range(self.X.shape[0]):
            self.assertTrue(np.array_equal(np.squeeze(np.reshape(orig_transformed[graph, ...], (-1, 1))),
                                           gembed[graph, ...]))

    def test_embedding_hope_2d(self):
        g_embedding = GraphEmbeddingHOPE(embedding_dimension=2)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertEqual((20, 20, 2), np.shape(gembed))

    def test_embedding_hope_4d(self):
        g_embedding = GraphEmbeddingHOPE(embedding_dimension=4)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertEqual((20, 20, 4), np.shape(gembed))



import numpy as np
from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor


class GraphConstructorPercentageWindow(GraphConstructor):
    _estimator_type = "transformer"

    def __init__(self,
                 percentage_upper: float = 50,
                 percentage_lower: float = 10,
                 retain_weights: float = 0,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for generating adjacency matrices
        from connectivity matrices. Selects the top x percent
        of connections and sets all other connections to zero


        Parameters
        ----------
        percentage_upper: float
            upper limit of the percentage window
        percentage_lower: float
            lower limit of the percentage window
        one_hot_nodes: int
            Whether to generate a one hot encoding of the nodes in the matrix (1) or not (0)
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int
            whether to perform a fisher transform of each matrix (1) or not (0)
        use_abs_fisher: int,default=0
            changes the values to absolute values. Is applied after fisher transform and before z-score transformation
        zscore: int,default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
        use_abs_zscore: int,default=0
            whether to use the absolute values of the z-score transformation or allow for negative values
        adjacency_axis: int
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data

        Example
        -------
        Use outside of a PHOTON pipeline

        ```python
        constructor = GraphConstructorPercentageWindow(percentage_upper=0.9,
                                                       percentage_lower=0.7
                                                       fisher_transform=1,
                                                       use_abs=1)
        ```

        Or as part of a pipeline

        ```python
        my_pipe.add(PipelineElement('GraphConstructorPercentageWindow',
                                    hyperparameters={'percentage_upper': 0.9, 'percentage_lower': 0.7}))
        ```
       """
        super(GraphConstructorPercentageWindow, self).__init__(one_hot_nodes=one_hot_nodes,
                                                               use_abs=use_abs,
                                                               fisher_transform=fisher_transform,
                                                               use_abs_fisher=use_abs_fisher,
                                                               zscore=zscore,
                                                               use_abs_zscore=use_abs_zscore,
                                                               adjacency_axis=adjacency_axis,
                                                               logs=logs)
        self.percentage_upper = percentage_upper
        self.percentage_lower = percentage_lower
        self.retain_weights = retain_weights
        if retain_weights not in [1, 0]:
            raise ValueError('retain_weights has to be in [1, 0]')

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Select percent strongest connections"""
        adj, feat = self.get_mtrx(X)
        # prepare matrices
        adj = self.prep_mtrx(adj)
        # get percent strongest connections
        adj = self.percentage_window(adj)
        # get feature matrix
        x_transformed = self.get_features(adj, feat)

        return x_transformed

    def percentage_window(self, adjacency: np.ndarray) -> np.ndarray:
        """Finds the x percent strongest connections"""
        for matrix in range(adjacency.shape[0]):
            thresh_upper = np.percentile(adjacency[matrix, :, :, :], self.percentage_upper)
            thresh_lower = np.percentile(adjacency[matrix, :, :, :], self.percentage_lower)
            adjacency[matrix, :, :, :][adjacency[matrix, :, :, :] <= thresh_lower] = 0
            adjacency[matrix, :, :, :][adjacency[matrix, :, :, :] >= thresh_upper] = 0
            if not self.retain_weights:
                adjacency[matrix, :, :, :][(adjacency[matrix, :, :, :] <= thresh_upper) &
                                           (adjacency[matrix, :, :, :] >= thresh_lower)] = 1

        return adjacency



import unittest
from scipy import sparse
import dgl
from photonai_graph.GraphConversions import dgl_to_sparse


class DglToSparseTest(unittest.TestCase):

    def setUp(self):
        # create dgl graph and nonsense input
        dgl_graph = dgl.DGLGraph()
        dgl_graph.add_nodes(3)
        dgl_graph.add_edges([0, 0, 1, 1, 2, 2], [1, 2, 0, 2, 0, 1])
        self.graphs = [dgl_graph] * 10
        self.edge_dict = {(1, 0): 1, (2, 0): 1}

    def test_nonsense_input(self):
        with self.assertRaises(Exception):
            dgl_to_sparse(self.edge_dict)

    def test_list_type(self):
        g = dgl_to_sparse(self.graphs)
        self.assertEqual(type(g), list)

    def test_output_format_csr(self):
        g = dgl_to_sparse(self.graphs, fmt="csr")
        self.assertEqual(type(g[0]), sparse.csr_matrix)

    def test_output_format_coo(self):
        g = dgl_to_sparse(self.graphs, fmt="coo")
        self.assertEqual(type(g[0]), sparse.coo_matrix)



from abc import ABC

import igraph
import numpy as np
import pandas as pd
import networkx as nx
from sklearn.base import BaseEstimator, TransformerMixin


class AbstractMeasureTransform(ABC, BaseEstimator, TransformerMixin):
    """
    Base class for measurement transforms
    """

    def __init__(self, graph_functions: dict = None):
        self.graph_functions = graph_functions

    def _shared_inner_transform(self, x_transformed):
        for c_measure in range(len(self.graph_functions)):
            expected_values = max([len(graph[c_measure]) for graph in x_transformed])
            for graph in x_transformed:
                if len(graph[c_measure]) < expected_values:
                    graph[c_measure] = [np.NAN] * expected_values

        return x_transformed

    @staticmethod
    def _shared_transform(x_transformed):
        for graph_idx in range(len(x_transformed)):
            g_m = list()
            for measure in x_transformed[graph_idx]:
                g_m.extend(measure)
            x_transformed[graph_idx] = g_m

        x_transformed = np.asarray(x_transformed)

        return x_transformed

    def handle_outputs(self, results, measure_list):
        # handle results
        if isinstance(results, dict):
            for rskey, rsval in results.items():
                self.handle_outputs(rsval, measure_list)
            return measure_list

        if isinstance(results, list):
            measure_list.extend(results)
            return measure_list

        # currently only networkx functions return tuples
        # The second return value can be discarded in these functions
        if isinstance(results, tuple):
            for result in results:
                self.handle_outputs(result, measure_list)
            return measure_list

        if isinstance(results, igraph.Graph):
            return measure_list

        if isinstance(results, nx.Graph):
            return measure_list

        measure_list.append(results)
        return measure_list

    def _shared_extraction(self, x_graphs, ids, path):
        x_transformed = self._inner_transform(x_graphs)

        measurements = []
        for graph, gid in zip(x_transformed, ids):
            for measurement_id, result in enumerate(graph):
                for res in result:
                    current_measurement = [gid, list(self.graph_functions.keys())[measurement_id], res]
                    measurements.append(current_measurement)

        df = pd.DataFrame(measurements)

        col_names = ["graph_id", "measure", "value"]

        df.to_csv(path_or_buf=path, header=col_names, index=None)

    def _inner_transform(self, X):
        """
        Inner transformation

        Parameters
        ----------
        X

        Returns
        -------

        """


import unittest
import numpy as np
import networkx as nx
from photonai_graph.GraphUtilities import check_asteroidal, check_asteroidal_boolean


class CheckAsteroidalTest(unittest.TestCase):

    def setUp(self):
        self.graph = nx.cycle_graph(20)
        self.graphs = [nx.cycle_graph(20)] * 10
        self.nonsense = np.random.rand(20, 20)
        self.path_graph = nx.path_graph(10)
        self.path_graphs = [nx.path_graph(10)] * 10

    def test_output_bool(self):
        g = check_asteroidal_boolean(self.graph)
        self.assertFalse(g)

    def test_output_nobool(self):
        g = check_asteroidal(self.path_graph)
        self.assertEqual(None, g)

    def test_output_list_bool(self):
        g = check_asteroidal_boolean(self.graphs)
        f_list = [False] * 10
        self.assertEqual(g, f_list)

    def test_output_list_nobool(self):
        g = check_asteroidal(self.path_graphs)
        v_list = [None] * 10
        self.assertEqual(v_list, g)

    def test_nonsense_bool(self):
        with self.assertRaises(ValueError):
            check_asteroidal_boolean(self.nonsense)

    def test_nonsense_nobool(self):
        with self.assertRaises(ValueError):
            check_asteroidal(self.nonsense)



import os
from typing import List
from functools import partial

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from tqdm.contrib.concurrent import process_map

from photonai_graph.Controllability.controllability_functions import modal_control, \
    average_control


class ControllabilityMeasureTransform(BaseEstimator, TransformerMixin):
    _estimator_type = "transformer"

    def __init__(self,
                 mod_control: int = 1,
                 ave_control: int = 1,
                 adjacency_axis: int = 0,
                 n_processes: int = 0,
                 logs: str = None):
        """
            Class for extraction of controllability measures. Allows
            for extraction of controllability measures.

            Parameters
            ----------
            mod_control: int,default = 1
                Whether to calculate nodewise modal controllability (1) or not (0).
            ave_control: int,default=1
                Whether to calculate nodewise average controllability (1) or not (0).
            adjacency_axis: int,default=0
                position of the adjacency matrix, default being zero
            n_processes: int,default=0
                Number of processes to use for multiprocessing
            logs: str,default=None
            """
        self.mod_control = mod_control
        self.ave_control = ave_control
        self.adjacency_axis = adjacency_axis
        self.n_processes = n_processes
        if logs:
            self.logs = logs
        else:
            self.logs = os.getcwd()
        if not self.mod_control and not self.ave_control:
            raise ValueError("You need to select either average, modal or both controllabilities."
                             "Please check your hyperparameter choices.")

    def fit(self, X, y):
        return self

    def __calculate_controllability(self, X, subj):
        vec = None
        if self.mod_control:
            vec = modal_control(X[subj, :, :, self.adjacency_axis])
        if self.ave_control:
            ac = average_control(X[subj, :, :, self.adjacency_axis])
            if vec is None:
                vec = ac
            else:
                vec = np.concatenate((vec, ac))
        return vec

    def transform(self, X, as_array: bool = True):

        controllabilities = []
        if self.n_processes == 0:
            for subj in range(X.shape[0]):
                vec = self.__calculate_controllability(X, subj)
                controllabilities.append(vec)
        else:
            fn = partial(self.__calculate_controllability, X=X)
            controllabilities = process_map(fn, np.arange(X.shape[0]))
        if not as_array:
            return controllabilities

        return np.asarray(controllabilities)

    def extract_measures(self,
                         X: np.ndarray = None,
                         path: str = None,
                         ids: List[int] = None,
                         node_list: List[str] = None):
        """Extract controllability measures and write them to a csv file

        Parameters
        ----------
        X: np.ndarray,default=None
            Input numpy array to transform
        path: str,default=None
            Output path for generated CSV
        ids: List[int],default=None
            List of ids for the graphs. If None the graphs are enumerated
        node_list: List[str],default=None
            List of names for the nodes. If None the nodes of the graphs are enumerated and entitled by the calculated
            controllability measure

        """

        # check if id list is correct length
        if ids is not None:
            if len(ids) != len(list(range(X.shape[0]))):
                raise ValueError("ID list does not match number of samples.")
        else:
            ids = list(range(X.shape[0]))

        # check if node list is correct length
        if node_list is not None:
            if len(node_list) != len(list(range(X.shape[1]))):
                raise ValueError("Node list does not match number of nodes.")
        else:
            node_list = list(range(X.shape[1]))
            for i in node_list:
                node_list[i] = "node_" + str(node_list[i])

        # base columns for every csv
        cols = ["ID"]

        controllabilities = self.transform(X, as_array=False)

        df = pd.DataFrame(controllabilities)
        df.insert(0, "ID", ids)

        if self.mod_control:
            cols += self.__add_prefix(node_list, "modal control ")
        if self.ave_control:
            cols += self.__add_prefix(node_list, "average control ")
        df.to_csv(path_or_buf=path, header=cols, index=False)

    @staticmethod
    def __add_prefix(x_list, prefix):
        """Adds a prefix to a list of strings"""
        y_list = x_list.copy()
        for i in range(len(y_list)):
            y_list[i] = prefix + y_list[i]

        return y_list



import unittest
import numpy as np
from photonai_graph.GraphUtilities import get_random_labels


class DrawConnectogramTests(unittest.TestCase):

    def setUp(self):
        self.np_control = np.random.rand(20)

    def test_type(self):
        y = get_random_labels(number_of_labels=20)
        self.assertEqual(type(y), np.ndarray)

    def test_shape(self):
        y = get_random_labels(number_of_labels=20)
        self.assertEqual(np.shape(y), (20,))

    def test_classification(self):
        y = get_random_labels(number_of_labels=20)
        self.assertEqual(len(np.unique(y)), 2)

    def test_regression(self):
        y = get_random_labels(l_type="regression", number_of_labels=20)
        self.assertGreaterEqual(len(np.unique(y)), 3)

    def test_value_error(self):
        with self.assertRaises(ValueError):
            get_random_labels(l_type="prediction", number_of_labels=20)



import unittest
import networkx as nx
from photonai_graph.GraphUtilities import pydot_to_nx


class PydotToNxTest(unittest.TestCase):

    def setUp(self):
        nx_graph = nx.erdos_renyi_graph(20, 0.3)
        py_graph = nx.nx_pydot.to_pydot(nx_graph)
        self.nx_graph = nx_graph
        self.graph_single = py_graph
        self.graph_list = [py_graph] * 3

    def test_reading_single(self):
        nx_graph = pydot_to_nx(self.graph_single)
        self.assertEqual(type(nx_graph), nx.classes.Graph)

    def test_reading_list(self):
        nx_graphs = pydot_to_nx(self.graph_list)
        self.assertEqual(type(nx_graphs), list)

    def test_reading_list_type(self):
        nx_graphs = pydot_to_nx(self.graph_list)
        self.assertEqual(type(nx_graphs[0]), nx.classes.Graph)

    def test_type_error(self):
        with self.assertRaises(TypeError):
            pydot_to_nx(self.nx_graph)



import unittest
import numpy as np

from photonai_graph.GraphKernels.grakel_adapter import GrakelAdapter
from photonai_graph.GraphKernels.GrakelTransformer import GrakelTransformer


class GrakelTransformerTest(unittest.TestCase):

    def setUp(self) -> None:
        mtrx = np.random.rand(20, 20, 20, 2)
        mtrx[:, :, :, 0][mtrx[:, :, :, 0] <= 0.9] = 0
        self.Xrandom4d = mtrx
        self.Xrandom_missfit = np.random.rand(20, 20, 5, 2)
        self.y = np.ones(20)

    def test_dummy_function(self):
        g_adapter = GrakelAdapter()
        x_trans = g_adapter.transform(self.Xrandom4d)
        self.assertEqual(type(x_trans), list)
        gt = GrakelTransformer(transformation="ShortestPath", with_labels=True)
        gt.fit(x_trans)
        res = gt.transform(x_trans)
        self.assertTrue(isinstance(res, np.ndarray))

    def test_unknown_function(self):
        with self.assertRaises(ValueError):
            GrakelTransformer(transformation="Dummy")

    def test_unfit_transformer(self):
        g_adapter = GrakelAdapter()
        x_trans = g_adapter.transform(self.Xrandom4d)
        gt = GrakelTransformer(transformation="ShortestPath")
        with self.assertRaises(ValueError):
            gt.transform(x_trans)



import unittest

import numpy as np
import networkx as nx
import igraph
from photonai_graph.Measures.MeasureAdapter import GraphMeasureAdapter
from photonai_graph.util import NetworkxGraphWrapper


class MeasureAdapterTests(unittest.TestCase):

    def setUp(self):
        # generate random graphs
        self.random_mtrx = np.random.rand(10, 20, 20, 2)
        self.random_igraph = [igraph.Graph.Erdos_Renyi(n=10, p=0.3) for _ in range(10)]
        self.random_networkx = [nx.erdos_renyi_graph(n=10, p=0.3) for _ in range(10)]
        # generate nonsense input
        self.edge_dict = {(1, 0): 1, (2, 0): 1, (3, 0): 1}
        self.y = np.random.rand(10)

    def test_np_to_igraph(self):
        adapter = GraphMeasureAdapter(output='igraph')
        adapter.fit(self.random_mtrx, self.y)
        g = adapter.transform(self.random_mtrx)
        self.assertEqual(type(g[0]), igraph.Graph)
        self.assertEqual(type(g), np.ndarray)

    def test_igraph_to_igraph(self):
        adapter = GraphMeasureAdapter(output='igraph')
        g = adapter.transform(self.random_igraph)
        self.assertEqual(type(g[0]), type(self.random_igraph[0]))

    def test_igraph_type_error(self):
        adapter = GraphMeasureAdapter(output='igraph')
        with self.assertRaises(TypeError):
            adapter.transform(self.edge_dict)

    def test_np_to_networkx(self):
        adapter = GraphMeasureAdapter(output='networkx')
        adapter.fit(self.random_mtrx, self.y)
        g = adapter.transform(self.random_mtrx)
        self.assertEqual(type(g[0]), NetworkxGraphWrapper)
        self.assertEqual(type(g), np.ndarray)

    def test_networkx_to_networkx(self):
        adapter = GraphMeasureAdapter(output='networkx')
        g = adapter.transform(self.random_networkx)
        self.assertEqual(type(g[0]), NetworkxGraphWrapper)

    def test_output_type_error(self):
        adapter = GraphMeasureAdapter(output='test')
        with self.assertRaises(NotImplementedError):
            adapter.transform(self.random_mtrx)



import unittest
import numpy as np
from photonai_graph.DynamicUtils.cofluct_functions import cofluct


class CofluctFunctionTests(unittest.TestCase):

    def setUp(self):
        self.X_test2d = np.random.rand(20, 2000)

    def test_cofluct(self):
        res = cofluct(self.X_test2d, (0, 1), return_mat=True)
        self.assertEqual(np.testing.assert_array_almost_equal(res, np.corrcoef(self.X_test2d)), None)

    def test_dict(self):
        res = cofluct(self.X_test2d, (0, 1), return_mat=False)
        self.assertEqual(len(res), 191)



from photonai_graph.GraphConstruction.graph_constructor_percentage_window import GraphConstructorPercentageWindow
from photonai_graph.GraphConstruction.graph_constructor_random_walks import GraphConstructorRandomWalks
from photonai_graph.GraphConstruction.graph_constructor_percentage import GraphConstructorPercentage
from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor
from photonai_graph.GraphConstruction.graph_constructor_knn import GraphConstructorKNN
from photonai_graph.GraphConstruction.graph_constructor_spatial import GraphConstructorSpatial
from photonai_graph.GraphConstruction.graph_constructor_threshold import GraphConstructorThreshold
from photonai_graph.GraphConstruction.graph_constructor_threshold_window import GraphConstructorThresholdWindow



import unittest
import networkx as nx
import numpy as np
import os
from photonai_graph.GraphUtilities import draw_connectograms


class DrawConnectogramTests(unittest.TestCase):

    def setUp(self):
        self.cyc_graph = nx.cycle_graph(20)
        self.cyc_graphs = [nx.cycle_graph(20)] * 5
        self.matrix = np.random.rand(20, 20, 20)
        self.ids = list(range(6, 11))
        self.ids_wrong = list(range(1, 11))
        self.directory = './'

    def test_drawing_single(self):
        draw_connectograms(self.cyc_graph, show=False)

    def test_drawing_list(self):
        draw_connectograms(self.cyc_graphs, show=False)

    def test_drawing_len_ids(self):
        output_format = ".svg"
        draw_connectograms(self.cyc_graphs, path=self.directory,
                           ids=self.ids, out_format=output_format, show=False)
        self.assertTrue(os.path.exists(self.directory + str(6) + output_format))
        [os.remove(os.path.join(self.directory, f"{cid}{output_format}")) for cid in self.ids]

    def test_drawing_len_ids_wrong(self):
        with self.assertRaises(Exception):
            draw_connectograms(self.cyc_graphs, path=self.directory,
                               ids=self.ids_wrong, out_format='.svg', show=False)

    def test_drawing_path(self):
        with self.assertRaises(Exception):
            draw_connectograms(self.cyc_graphs, ids=self.ids, out_format='.svg', show=False)

    def test_drawing_outformat(self):
        with self.assertRaises(Exception):
            draw_connectograms(self.cyc_graphs, path=self.directory,
                               ids=self.ids, show=False)

    def test_save_no_ids(self):
        output_format = ".png"
        draw_connectograms(self.cyc_graphs, path=self.directory, out_format=output_format, show=False)
        self.assertTrue(os.path.exists(self.directory + str(0) + output_format))
        [os.remove(os.path.join(self.directory, f"{cid}{output_format}")) for cid, _ in enumerate(self.cyc_graphs)]

    def test_format(self):
        with self.assertRaises(TypeError):
            draw_connectograms(self.matrix, show=False)



import unittest
import numpy as np
from photonai_graph.GraphConstruction.graph_constructor_threshold_window import GraphConstructorThresholdWindow


class ThresholdWindowTests(unittest.TestCase):

    def setUp(self):
        self.X4d = np.ones((20, 20, 20, 2))
        self.X3d = np.ones((20, 20, 20))
        self.Xrandom4d = np.random.rand(20, 20, 20, 2)
        test_array = np.reshape(np.arange(1, 101, 1), (-1, 10, 10))
        self.Xtest4d = np.repeat(test_array, 10, axis=0)[..., np.newaxis]
        self.y = np.ones((20))

    def test_treshold_window_mean(self):
        g_constr = GraphConstructorThresholdWindow()
        g_constr.fit(self.X4d, self.y)
        trans = g_constr.transform(self.X4d)
        self.assertEqual(np.shape(trans), (20, 20, 20, 3))

    def test_threshold_window_retain_weights_none(self):
        g_constr = GraphConstructorThresholdWindow(retain_weights=0)
        g_constr.fit(self.Xrandom4d, self.y)
        trans = g_constr.transform(self.Xrandom4d)
        self.assertEqual(len(np.unique(trans[:, :, :, 0])), 2)

    def test_threshold_window_retain_weights(self):
        # ensure that retain weights leads to a diverse set of weights (more than 0 and 1)
        g_constr = GraphConstructorThresholdWindow(retain_weights=1)
        g_constr.fit(self.Xrandom4d, self.y)
        trans = g_constr.transform(self.Xrandom4d)
        self.assertEqual(np.min(trans[:, :, :, 0]), 0)
        self.assertGreaterEqual(len(np.unique(trans[:, :, :, 0])), 2)

    def test_threshold_window_retain_weights_error(self):
        # ensure a Value error is being raised when retain weights is neither 0 or 1
        with self.assertRaises(ValueError):
            g_constr = GraphConstructorThresholdWindow(retain_weights=0.5)
            g_constr.fit(self.Xrandom4d, self.y)
            g_constr.transform(self.Xrandom4d)

    def test_threshold_contains(self):
        # ensure that the threshold actually picks the right rows of the matrix
        g_constr = GraphConstructorThresholdWindow(threshold_upper=80,
                                                   threshold_lower=70,
                                                   retain_weights=1)
        g_constr.fit(self.Xtest4d, self.y)
        trans = g_constr.transform(self.Xtest4d)
        expected_elements = [70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80]
        for element in expected_elements:
            self.assertIn(element, trans[0, :, :, 0])

    def test_threshold_not_contains(self):
        # ensure that the threshold actually picks the right rows of the matrix
        g_constr = GraphConstructorThresholdWindow(threshold_upper=80,
                                                   threshold_lower=70,
                                                   retain_weights=1)
        g_constr.fit(self.Xtest4d, self.y)
        trans = g_constr.transform(self.Xtest4d)
        expected_elements = [90, 91, 92, 93, 94, 95, 96, 97, 98, 99]
        for element in expected_elements:
            self.assertNotIn(element, trans[0, :, :, 0])



from typing import List

import numpy as np
try:
    import dgl
    import torch.nn as nn
    from dgl.nn.pytorch import GATConv
except ImportError:
    pass

from photonai_graph.NeuralNets.dgl_base import DGLClassifierBaseModel, DGLRegressorBaseModel


class GATModel(nn.Module):
    def __init__(self, in_dim, hidden_dim, num_heads, n_classes, hidden_layers,
                 agg_mode, allow_zero_in_degree):
        super(GATModel, self).__init__()

        self.agg_mode = agg_mode
        self.layers = nn.ModuleList()
        # append gat layer according to inputs
        self.layers.append(GATConv(in_dim, hidden_dim, num_heads[0],
                                   allow_zero_in_degree=allow_zero_in_degree))
        # hidden layers
        for layer in range(1, hidden_layers):
            self.layers.append(GATConv(
                hidden_dim, hidden_dim, num_heads[layer],
                allow_zero_in_degree=allow_zero_in_degree))
        # output layer
        self.classify = nn.Linear(hidden_dim, n_classes)

    def forward(self, bg):
        # For undirected graphs, in_degree is the same as
        # out_degree.
        h = bg.in_degrees().view(-1, 1).float()
        for i, gnn in enumerate(self.layers):
            h = gnn(bg, h)
            if self.agg_mode == 'flatten':
                h = h.flatten(1)
            else:
                h = h.mean(1)

        bg.ndata['h'] = h
        hg = dgl.mean_nodes(bg, 'h')
        return self.classify(hg)


class GATClassifierModel(DGLClassifierBaseModel):

    def __init__(self,
                 in_dim: int = 1,
                 hidden_layers: int = 2,
                 hidden_dim: int = 256,
                 heads: List = None,
                 agg_mode="mean",
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 early_stopping: bool = False,
                 es_patience: int = 10,
                 es_tolerance: int = 9,
                 es_delta: float = 0,
                 verbose: bool = False,
                 logs: str = '',
                 **kwargs):
        """
            Graph Attention Network for graph classification. GAT Layers
            are modeled after Veličković et al., 2018.
            Implementation based on dgl & pytorch.


            Parameters
            ----------
            in_dim: int,default=1
                input dimension
            hidden_layers: int,default=2
                number of hidden layers used by the model
            hidden_dim: int,default=256
                dimensions in the hidden layers
            heads: list,default=None
                list with number of heads per hidden layer
            validation_score: bool,default=False
                If true the input data is split into train and test (90%/10%).
                The testset is then used to get validation results during training
            verbose: bool,default=False
                If true verbose output is generated
            agg_mode: str, default='mean'
                aggregation mode for the graph convolutional layers

        """
        super(GATClassifierModel, self).__init__(nn_epochs=nn_epochs,
                                                 learning_rate=learning_rate,
                                                 batch_size=batch_size,
                                                 adjacency_axis=adjacency_axis,
                                                 feature_axis=feature_axis,
                                                 add_self_loops=add_self_loops,
                                                 allow_zero_in_degree=allow_zero_in_degree,
                                                 validation_score=validation_score,
                                                 early_stopping=early_stopping,
                                                 es_patience=es_patience,
                                                 es_tolerance=es_tolerance,
                                                 es_delta=es_delta,
                                                 verbose=verbose,
                                                 logs=logs,
                                                 **kwargs)
        if heads is None:
            heads = [2, 2]
            # Todo: if heads is not length of hidden layers +1 (bc of the first layer)
        self.in_dim = in_dim
        self.hidden_layers = hidden_layers
        self.hidden_dim = hidden_dim
        self.heads = heads
        self.agg_mode = agg_mode

    def _init_model(self, X=None, y=None):
        self.model = GATModel(self.in_dim, self.hidden_dim, self.heads,
                              len(np.unique(y)), self.hidden_layers, self.agg_mode,
                              allow_zero_in_degree=self.allow_zero_in_degree)


class GATRegressorModel(DGLRegressorBaseModel):

    def __init__(self,
                 in_dim: int = 1,
                 hidden_layers: int = 2,
                 hidden_dim: int = 256,
                 heads: List = None,
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 logs: str = None,
                 validation_score: bool = False,
                 early_stopping: bool = False,
                 es_patience: int = 10,
                 es_tolerance: int = 9,
                 es_delta: float = 0,
                 verbose: bool = False,
                 agg_mode: str = None,
                 **kwargs):
        """
            Graph Attention Network for graph regression. GAT Layers
            are modeled after Veličković et al., 2018. Implementation
            based on dgl & pytorch.


            Parameters
            ----------
            in_dim: int,default=1
                input dimension
            hidden_layers: int,default=2
                number of hidden layers used by the model
            hidden_dim: int,default=256
                dimensions in the hidden layers
            heads: list,default=None
                list with number of heads per hidden layer
            validation_score: bool,default=False
                If true the input data is split into train and test (90%/10%).
                The testset is then used to get validation results during training
            verbose: bool,default=False
                If true verbose output is generated
        """
        super(GATRegressorModel, self).__init__(nn_epochs=nn_epochs,
                                                learning_rate=learning_rate,
                                                batch_size=batch_size,
                                                adjacency_axis=adjacency_axis,
                                                feature_axis=feature_axis,
                                                add_self_loops=add_self_loops,
                                                allow_zero_in_degree=allow_zero_in_degree,
                                                validation_score=validation_score,
                                                early_stopping=early_stopping,
                                                es_patience=es_patience,
                                                es_tolerance=es_tolerance,
                                                es_delta=es_delta,
                                                verbose=verbose,
                                                logs=logs,
                                                **kwargs)
        if heads is None:
            heads = [2, 2]
            # Todo: if heads is not length of hidden layers +1 (bc of the first layer)
        self.in_dim = in_dim
        self.hidden_layers = hidden_layers
        self.hidden_dim = hidden_dim
        self.heads = heads
        self.agg_mode = agg_mode

    def _init_model(self, X=None, y=None):
        self.model = GATModel(self.in_dim, self.hidden_dim, self.heads, 1, self.hidden_layers,
                              allow_zero_in_degree=self.allow_zero_in_degree, agg_mode=self.agg_mode).float()



import numpy as np

from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor


class GraphConstructorThresholdWindow(GraphConstructor):
    _estimator_type = "transformer"

    def __init__(self,
                 threshold_upper: float = 1,
                 threshold_lower: float = 0.8,
                 retain_weights: float = 0,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for generating adjacency matrices
        from connectivity matrices. Thresholds matrix based
        on a chosen threshold window.


        Parameters
        ----------
        threshold_upper: float
            upper limit of the threshold window
        threshold_lower: float
            lower limit of the threshold window
        retain_weights: int
            whether to retain weight values or not
        one_hot_nodes: int
            Whether to generate a one hot encoding of the nodes in the matrix.
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int
            Perform a fisher transform of each matrix. No (0) or Yes (1)
        use_abs_fisher: int
            Changes the values to absolute values. Is applied after fisher transform and before
            z-score transformation
        zscore: int, default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
            eval_final_perfomance is set to True
        use_abs_zscore: int, default=0
            whether to use the absolute values of the z-score transformation or allow for negative
            values. Applied after fisher transform, use_abs and zscore
        adjacency_axis: int
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data

        Example
        -------
        Use outside of a PHOTON pipeline

        ```python
        constructor = GraphConstructorThresholdWindow(threshold_upper=0.7,
                                                      threshold_lower=0.3,
                                                      use_abs=1)
        ```

        Or as part of a pipeline

        ```python
        my_pipe.add(PipelineElement('GraphConstructorThresholdWindow',
                                    hyperparameters={'threshold_upper': 0.7, 'threshold_lower': 0.3}))
        ```
       """
        super(GraphConstructorThresholdWindow, self).__init__(one_hot_nodes=one_hot_nodes,
                                                              use_abs=use_abs,
                                                              fisher_transform=fisher_transform,
                                                              use_abs_fisher=use_abs_fisher,
                                                              zscore=zscore,
                                                              use_abs_zscore=use_abs_zscore,
                                                              adjacency_axis=adjacency_axis,
                                                              logs=logs)
        self.threshold_upper = threshold_upper
        self.threshold_lower = threshold_lower
        self.retain_weights = retain_weights
        if retain_weights not in [0, 1]:
            raise ValueError("retain_weights has to be in [0, 1]")

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform input matrices accordingly"""
        adj, feat = self.get_mtrx(X)
        # do preparatory matrix transformations
        adj = self.prep_mtrx(adj)
        # threshold matrix
        adj = self.threshold_window(adj)
        # get feature matrix
        transformed = self.get_features(adj, feat)

        return transformed

    def threshold_window(self, adjacency: np.ndarray) -> np.ndarray:
        """Threshold matrix"""
        adjacency[adjacency < self.threshold_lower] = 0
        adjacency[adjacency > self.threshold_upper] = 0
        if not self.retain_weights:
            adjacency[(adjacency < self.threshold_upper) & (adjacency >= self.threshold_lower)] = 1
        return adjacency



import unittest
import numpy as np
from photonai_graph.GraphEmbedding.graph_embedding_laplacian_eigenmaps import GraphEmbeddingLaplacianEigenmaps
from photonai_graph.GraphUtilities import get_random_connectivity_data, get_random_labels


class LaplacianEigenmapsTest(unittest.TestCase):

    def setUp(self):
        mtrx = np.random.rand(20, 20, 20, 2)
        mtrx[:, :, :, 0][mtrx[:, :, :, 0] <= 0.9] = 0
        self.X = mtrx
        self.y = np.ones(20)

    def test_embedding_laplacian_eigenmaps(self):
        g_embedding = GraphEmbeddingLaplacianEigenmaps(embedding_dimension=1)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertEqual((20, 20), np.shape(gembed))

    def test_embedding_laplacian_eigenmaps_complex(self):
        g_embedding = GraphEmbeddingLaplacianEigenmaps(embedding_dimension=1)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertFalse(np.iscomplex(gembed).any())

    def test_embedding_laplacian_eigenmaps_complex_random(self):
        X = get_random_connectivity_data(number_of_nodes=50, number_of_individuals=100)
        y = get_random_labels(l_type="regression", number_of_labels=100)
        g_embedding = GraphEmbeddingLaplacianEigenmaps()
        g_embedding.fit(X, y)
        embedded = g_embedding.transform(X)
        self.assertFalse(np.iscomplex(embedded).any())

    def test_embedding_laplacian_eigenmaps_3d(self):
        g_embedding = GraphEmbeddingLaplacianEigenmaps(embedding_dimension=3)
        g_embedding.fit(self.X, self.y)
        gembed = g_embedding.transform(self.X)
        self.assertEqual((20, 20, 3), np.shape(gembed))



import unittest
import numpy as np
import networkx as nx
from photonai_graph.GraphUtilities import visualize_networkx


class VisualizeNetworkxTest(unittest.TestCase):

    def setUp(self):
        self.matrix = np.random.rand(20, 20)
        self.graph = nx.cycle_graph(20)
        self.graphs = [nx.erdos_renyi_graph(20, 0.3)] * 20

    def test_plot_single(self):
        # todo: fix plotting tests
        # visualize_networkx(self.graph, show=False)
        pass

    def test_plot_list(self):
        # todo: Fix plotting tests
        # visualize_networkx(self.graphs, show=False)
        pass

    def test_value_error(self):
        with self.assertRaises(ValueError):
            visualize_networkx(self.matrix, show=False)



import unittest
import numpy as np
from photonai_graph.GraphUtilities import individual_fishertransform


class IndividualFishertransformTest(unittest.TestCase):

    def setUp(self):
        self.m3d = np.random.rand(10, 20, 20)
        self.m4d = np.ones((10, 20, 20, 2))
        self.m4dnan = np.empty((10, 20, 20, 2)) * np.nan

    def test_fishertransform_3d(self):
        with self.assertRaises(ValueError):
            individual_fishertransform(self.m3d)

    def test_ztransform_4d(self):
        trans_m4d = individual_fishertransform(self.m4d)
        self.assertEqual(np.shape(trans_m4d), np.shape(self.m4d[..., 0, np.newaxis]))

    def test_fishertransform_nan(self):
        with self.assertRaises(ValueError):
            individual_fishertransform(self.m4dnan)



import unittest
import numpy as np
from photonai_graph.GraphConstruction.graph_constructor_knn import GraphConstructorKNN


class KNNTests(unittest.TestCase):

    def setUp(self):
        self.X4d = np.ones((20, 20, 20, 2))
        self.Xrandom4d = np.random.rand(20, 20, 20, 2)
        test_array4d = np.array([[1, 1, 1, 0.3, 0.2, 0.4],
                                 [1, 1, 1, 0.2, 0.1, 0.1],
                                 [1, 1, 1, 0.4, 0.2, 0.2],
                                 [0.2, 0.3, 0.1, 1, 1, 1],
                                 [0.3, 0.4, 0.3, 1, 1, 1],
                                 [0.1, 0.4, 0.3, 1, 1, 1]])
        test_array4d = test_array4d[np.newaxis, :, :, np.newaxis]
        self.Xtest4d = np.repeat(test_array4d, 10, axis=0)
        self.y = np.ones((20))

    def test_knn_individual(self):
        g_constr = GraphConstructorKNN()
        g_constr.fit(self.Xrandom4d, self.y)
        trans = g_constr.transform(self.Xrandom4d)
        self.assertEqual(np.shape(trans), (20, 20, 20, 3))

    def test_knn_assert_failure(self):
        with self.assertRaises(AssertionError):
            g_constr = GraphConstructorKNN()
            g_constr.fit(self.X4d, self.y)
            g_constr.transform(self.X4d)

    def test_knn_mechanism(self):
        g_constr = GraphConstructorKNN(k_distance=2)
        g_constr.fit(self.Xtest4d, self.y)
        trans = g_constr.transform(self.Xtest4d)
        bool_mask = [[True, False, False, True, True, True],
                     [False, True, False, True, True, True],
                     [False, False, True, True, True, True],
                     [True, True, True, True, False, False],
                     [True, True, True, False, True, False],
                     [True, True, True, False, False, True]]
        bool_array = trans[0, :, :, 0] == 0
        comp = bool_mask == bool_array
        comp = comp.all()
        self.assertTrue(comp)



import networkx as nx
import seaborn as sns
import pandas as pd
from matplotlib import pyplot as plt


def get_random_graphs(num_graphs=500, num_nodes=20, edge_prob=0.3):
    # returns a list of random networkx graphs
    graph_list = []
    for _ in range(num_graphs):
        graph = nx.fast_gnp_random_graph(num_nodes, edge_prob)
        graph_list.append(graph)

    return graph_list


def get_sw_graph(num_graphs=500, num_nodes=20, knn=5, edge_rew=0.2):
    # returns a list of small world networkx graphs
    graph_list = []
    for _ in range(num_graphs):
        graph = nx.watts_strogatz_graph(num_nodes, knn, edge_rew)
        graph_list.append(graph)

    return graph_list


def plot_nx_edge_count(graphs1, graphs2, label1="graphs_1", label2="graphs_2"):
    # plot the distribution of edges for 2 sets of graphs
    edges1 = []
    edges2 = []

    for graph in graphs1:
        edge_count = graph.number_of_edges()
        edges1.append(edge_count)

    for graph in graphs2:
        edge_count = graph.number_of_edges()
        edges2.append(edge_count)

    labels = ([label1] * len(edges1)) + ([label2] * len(edges2))
    edges = edges1 + edges2

    df = pd.DataFrame(list(zip(edges, labels)), columns=["edge_count", "Group"])

    sns.displot(df, x="edge_count", hue="Group", kind="kde", fill=True)
    plt.show()



from photonai_graph.DynamicUtils.cofluct_functions import cofluct
from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import os


class CofluctTransform(BaseEstimator, TransformerMixin):
    _estimator_type = "transformer"

    def __init__(self,
                 quantiles: tuple = (0, 1),
                 return_mat: bool = True,
                 adjacency_axis: int = 0,
                 logs=None):
        """
        Class for calculating time series co-activation. Based on
        Esfahlani et al, 2020.

        Parameters
        ----------
        quantiles: tuple,default=(0, 1)
            lower and upper bound of connection strength quantile to look at.
        return_mat: bool,default=True
            Whether to return matrix (True) or vector (False).
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero

        Example
        -------
        ```python
            transformer = CofluctTransform(quantiles=(0.95, 1), return_mat=True)
        ```
        """
        self.quantiles = quantiles
        self.return_mat = return_mat
        self.adjacency_axis = adjacency_axis
        if logs:
            self.logs = logs
        else:
            self.logs = os.getcwd()

    def fit(self, X, y):
        return self

    def transform(self, X):

        # check if matrix is 3d or 4d
        if np.ndim(X) == 3:
            mtrx = X.copy()
        elif np.ndim(X) == 4:
            mtrx = X[:, :, :, self.adjacency_axis]
        else:
            raise TypeError("Matrix needs to 3d or 4d."
                            "Individuals x Nodes x Timepoints (x Modalities).")

        # calculate cofluct for each matrix
        cfls = []

        for m in range(mtrx.shape[0]):
            ser = cofluct(mtrx[m, :, :], self.quantiles, self.return_mat)
            cfls.append(ser)

        np.asarray(cfls)

        return cfls



import numpy as np
from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor


class GraphConstructorPercentage(GraphConstructor):
    _estimator_type = "transformer"

    def __init__(self,
                 percentage: float = 80,
                 retain_weights: float = 0,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for generating adjacency matrices
        from connectivity matrices. Selects the top x percent
        of connections and sets all other connections to zero


        Parameters
        ----------
        percentage: float
            value of percent of connections to discard. A value of 90 keeps only the top 10%
        retain_weights: float
            whether to retain weight values or not
        one_hot_nodes: int, default=0
            Whether to generate a one hot encoding of the nodes in the matrix (1) or not (0)
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int, default=0
            whether to perform a fisher transform of each matrix (1) or not (0)
        use_abs_fisher: int, default=0
            changes the values to absolute values. Is applied after fisher transform and before z-score transformation
        zscore: int, default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
        use_abs_zscore: int, default=0
            whether to use the absolute values of the z-score transformation or allow for negative values
        adjacency_axis: int, default=0
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data

        Example
        -------
        Use outside of a PHOTON pipeline

        ```python
        constructor = GraphConstructorPercentage(percentage=0.9,
                                                 fisher_transform=1,
                                                 use_abs=1)
        ```

        Or as part of a pipeline

        ```python
        my_pipe.add(PipelineElement('GraphConstructorPercentage',
                                    hyperparameters={'percentage': 0.9}))
        ```


       """
        super(GraphConstructorPercentage, self).__init__(one_hot_nodes=one_hot_nodes,
                                                         use_abs=use_abs,
                                                         fisher_transform=fisher_transform,
                                                         use_abs_fisher=use_abs_fisher,
                                                         zscore=zscore,
                                                         use_abs_zscore=use_abs_zscore,
                                                         adjacency_axis=adjacency_axis,
                                                         logs=logs)
        self.percentage = percentage
        self.retain_weights = retain_weights
        if retain_weights not in [0, 1]:
            raise ValueError("retain_weights has to be in [0, 1]")

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Select percent strongest connections"""
        adj, feat = self.get_mtrx(X)
        # prepare matrices
        adj = self.prep_mtrx(adj)
        # get percent strongest connections
        adj = self.percent_strongest(adj)
        # get feature matrix
        x_transformed = self.get_features(adj, feat)

        return x_transformed

    def percent_strongest(self, adjacency: np.ndarray) -> np.ndarray:
        """Finds the x percent strongest connections"""
        for matrix in range(adjacency.shape[0]):
            thresh = np.percentile(adjacency[matrix, ...], self.percentage)
            if np.isnan(thresh):
                raise ValueError('Input contains NaN -> can not threshold')
            else:
                adjacency[matrix, ...][adjacency[matrix, ...] < thresh] = 0
                if not self.retain_weights:
                    adjacency[matrix, ...][adjacency[matrix, ...] >= thresh] = 1

        return adjacency



import unittest
import dgl
import networkx as nx
from scipy import sparse
import numpy as np
from photonai_graph.GraphConversions import check_dgl


class DglToNetworkxTest(unittest.TestCase):

    def setUp(self):
        # create dgl graphs
        dgl_graph = dgl.DGLGraph()
        dgl_graph.add_nodes(3)
        dgl_graph.add_edges([0, 0, 1, 1, 2, 2], [1, 2, 0, 2, 0, 1])
        self.dgl_graph_list = [dgl_graph] * 10
        # create networkx graphs
        nx_graph = nx.cycle_graph(5)
        self.nx_graph_list = [nx_graph] * 10
        # create scipy matrix
        sparse_matrix = sparse.csr_matrix([[0, 1, 1],
                                           [1, 0, 1],
                                           [1, 1, 0]])
        self.sp_matrix_list = [sparse_matrix] * 10
        # create numpy matrix
        array = np.array([[0, 1, 1],
                          [1, 0, 1],
                          [1, 1, 0]])
        m4d_array = array[np.newaxis, :, :, np.newaxis]
        individuals_array = np.repeat(m4d_array, 5, axis=0)
        modality_array = np.repeat(individuals_array, 2, axis=3)
        self.np_4d_array = modality_array
        self.np_5d_array = modality_array[:, :, :, :, np.newaxis]
        # create nonsense input
        self.edge_dict = {(1, 0): 1, (2, 0): 1}
        self.np_list = [np.ones((3, 3))] * 10

    def test_nonsense_input(self):
        with self.assertRaises(TypeError):
            check_dgl(self.edge_dict)

    def test_list_dgl(self):
        g = check_dgl(self.dgl_graph_list)
        self.assertEqual(len(g), 10)

    def test_dgl_output_num_nodes(self):
        g = check_dgl(self.dgl_graph_list)
        self.assertEqual(g[0].number_of_nodes(), 3)

    def test_dgl_output_num_edges(self):
        g = check_dgl(self.dgl_graph_list)
        self.assertEqual(g[0].number_of_edges(), 6)

    def test_numpy_list_exception(self):
        with self.assertRaises(Exception):
            check_dgl(self.np_list)

    def test_list_np4d(self):
        g = check_dgl(self.np_4d_array, adjacency_axis=0, feature_axis=1)
        self.assertEqual(len(g), 5)

    def test_np4d_conversion(self):
        g = check_dgl(self.np_4d_array, adjacency_axis=0, feature_axis=1)
        self.assertEqual(type(g[0]), dgl.DGLGraph)

    def test_np4d_output_num_nodes(self):
        g = check_dgl(self.np_4d_array, adjacency_axis=0, feature_axis=1)
        self.assertEqual(g[0].number_of_nodes(), 3)

    def test_np4d_output_num_edges(self):
        g = check_dgl(self.np_4d_array, adjacency_axis=0, feature_axis=1)
        self.assertEqual(g[0].number_of_edges(), 6)

    def test_np5d_error(self):
        with self.assertRaises(ValueError):
            check_dgl(self.np_5d_array, adjacency_axis=0, feature_axis=1)



import unittest
from tempfile import mkstemp

import numpy as np
import igraph
import os
from photonai_graph.Measures.IgraphMeasureTransform import IgraphMeasureTransform


class GraphMeasureTransformTests(unittest.TestCase):

    def setUp(self):
        self.er_graph = [igraph.Graph.Erdos_Renyi(n=10, p=0.5)] * 5
        self.y = [1, 1, 1, 0, 0]
        self.ids = [1, 2, 3, 4, 5]
        # generate nonsense input
        self.edge_dict = {(1, 0): 1, (2, 0): 1, (3, 0): 1}
        self.string_list = ['test'] * 5

    def test_default_measures(self):
        g_transform = IgraphMeasureTransform()
        g_transform.fit(self.er_graph, self.y)
        measures = g_transform.transform(self.er_graph)
        self.assertEqual(measures.shape, (5, 20))

    def test_transform_nonsense_input(self):
        g_transform = IgraphMeasureTransform()
        with self.assertRaises(TypeError):
            g_transform.fit(self.edge_dict, self.y)
            g_transform.transform(self.edge_dict)

    def test_transform_nonsense_list(self):
        g_tranform = IgraphMeasureTransform()
        with self.assertRaises(TypeError):
            g_tranform.transform(self.string_list)

    def test_transform_own_measures_parallel(self):
        g_transform = IgraphMeasureTransform(n_processes=2)
        g_transform.fit(self.er_graph, self.y)
        measures = g_transform.transform(self.er_graph)
        self.assertEqual(measures.shape, (5, 20))

    def test_transform_check_output_type(self):
        g_transform = IgraphMeasureTransform()
        g_transform.fit(self.er_graph, self.y)
        measures = g_transform.transform(self.er_graph)
        self.assertEqual(type(measures), np.ndarray)

    def test_transform_measure_single(self):
        g_transform = IgraphMeasureTransform(graph_functions={"radius": {}})
        g_transform.fit(self.er_graph, self.y)
        measures = g_transform.transform(self.er_graph)
        self.assertEqual(measures.shape, (5, 1))

    def test_transform_measure_multiple(self):
        g_transform = IgraphMeasureTransform(graph_functions={"eigenvector_centrality": {}})
        g_transform.fit(self.er_graph, self.y)
        measures = g_transform.transform(self.er_graph)
        self.assertEqual(measures.shape, (5, 10))

    def test_extract_nonsense_input_no_ids(self):
        g_transform = IgraphMeasureTransform()
        with self.assertRaises(ValueError):
            g_transform.fit(self.edge_dict, self.y)
            g_transform.extract_measures(self.edge_dict)

    def test_extract_default_measures(self):
        _, path = mkstemp('.csv')
        g_transform = IgraphMeasureTransform()
        g_transform.extract_measures(self.er_graph, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_extract_own_measures(self):
        _, path = mkstemp('.csv')
        g_transform = IgraphMeasureTransform(graph_functions={"eigenvector_centrality": {}})
        g_transform.extract_measures(self.er_graph, path, self.ids)
        self.assertTrue(os.path.exists(path))
        os.remove(path)



from sklearn.base import BaseEstimator, ClassifierMixin
from photonai_graph.util import assert_imported
import numpy as np
try:
    from grakel import graph
    import grakel
except ImportError:  # pragma: no cover
    pass


class GrakelAdapter(BaseEstimator, ClassifierMixin):

    def __init__(self,
                 input_type: str = "dense",
                 node_labels: dict = None,
                 edge_labels: dict = None,
                 node_feature_construction: str = "mean",
                 feature_axis: int = 1,
                 adjacency_axis: int = 0):
        """
        A transformer class for transforming graphs into Grakel format.


        Parameters
        ----------
        input_type: str,default="dense"
            the type of the input, dense or networkx
        node_labels: list,default=None
            a list of the node labels, needed for shortest path kernel
        edge_labels: dict,default=None
            list of edge labels if graphs are constructed from networkx graphs
        node_feature_construction: str,default="mean"
            mode of feature construction for graphs constructed from adjacency matrices. "mean" takes the mean of the
            nodes edge weights, "sum" takes the sum of the nodes edge weights, "degree" takes the node degree,
             and "features" takes the nodes features as supplied in the feature matrix.
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero
        feature_axis: int,default=1
            position of the feature axis, default being 1


        Example
        ----------
        ```python
        adapter = GrakelAdapter(input_type="dense", node_feature_construction="features")
        ```
        """
        self.input_type = input_type
        self.node_labels = node_labels
        self.edge_labels = edge_labels
        self.node_feature_construction = node_feature_construction
        self.feature_axis = feature_axis
        self.adjacency_axis = adjacency_axis
        assert_imported(["grakel"])
        if input_type not in ['networkx', 'dense']:
            raise ValueError("Only networkx or dense conversions are supported.")

    def fit(self, X, y):
        return self

    def transform(self, X):
        """sklearn compatible graph conversion"""
        if self.input_type == "dense":
            node_features = self.construct_node_features(X)
            edge_features = self.construct_edge_features(X)
        else:
            node_features = None
            edge_features = None
        transformed = self.convert_grakel(X, self.input_type, node_features, edge_features, self.adjacency_axis)

        return transformed

    @staticmethod
    def convert_grakel(graphs, in_format, node_labels, edge_features, adjacency_axis):
        """convert graphs into grakel format"""
        g_trans = []
        if in_format == "dense":
            for g in range(graphs.shape[0]):
                conv_g = graph.Graph(graphs[g, :, :, adjacency_axis], node_labels=node_labels[g],
                                     edge_labels=edge_features[g])
                g_trans.append(conv_g)

        if in_format == "networkx":
            g_trans = grakel.graph_from_networkx(graphs)

        return g_trans

    def construct_node_features(self, matrices):
        """construct node features from the feature matrix"""
        label_list = []
        for mtrx in range(matrices.shape[0]):
            feat = self.get_dense_feature(matrices[mtrx, :, :, :], adjacency_axis=self.adjacency_axis,
                                          feature_axis=self.feature_axis, aggregation=self.node_feature_construction)
            label_list.append(feat)

        return label_list

    def construct_edge_features(self, matrices):
        """construct edge features from the feature or adjacency matrix"""
        label_list = []
        for mtrx in range(matrices.shape[0]):
            feat = self.get_dense_edge_features(matrices[mtrx, :, :, :], adjacency_axis=self.adjacency_axis,
                                                feature_axis=self.feature_axis)
            label_list.append(feat)

        return label_list

    @staticmethod
    def get_dense_edge_features(matrix, adjacency_axis, feature_axis):
        """returns the features for an edge label dictionary
            Parameters
            ---------
            matrix: np.matrix/np.ndarray
                feature matrix
            adjacency_axis: int
                position of the adjacency matrix
            feature_axis: int
                position of the feature matrix
        """
        edge_feat = {}
        for index, value in np.ndenumerate(matrix[:, :, adjacency_axis]):
            conn_key = (str(index[0]), str(index[1]))
            key_val = {conn_key: value}
            edge_feat.update(key_val)
        return edge_feat

    @staticmethod
    def get_dense_feature(matrix, adjacency_axis, feature_axis, aggregation="sum"):
        """returns the features for a networkx graph
            Parameters
            ---------
            matrix: np.matrix/np.ndarray
                feature matrix
            adjacency_axis: int
                position of the adjacency matrix
            feature_axis: int
                position of the feature matrix
            aggregation:
                method of feature construction, sum gives a row-wise sum,
                "mean" gives a row-wise mean, "node_degree" give a row-wise node-degree,
                features returns the entire row as the feature vector
        """
        if aggregation == "sum":
            features = np.sum(matrix[:, :, feature_axis], axis=1)
            features = features.tolist()
        elif aggregation == "mean":
            features = (np.sum(matrix[:, :, feature_axis], axis=1)) / matrix.shape[0]
            features = features.tolist()
        elif aggregation == "node_degree":
            features = np.count_nonzero(matrix[:, :, adjacency_axis], axis=1, keepdims=False)
            features = features.tolist()
        elif aggregation == "features":
            features = matrix[:, :, feature_axis]
            features = features.reshape((features.shape[0], -1))
            features = features.tolist()
        else:
            raise KeyError('Only sum, mean, node_degree and all features are supported')

        features = dict(enumerate(features, 0))

        return features



from sklearn.base import BaseEstimator, TransformerMixin
from abc import ABC, abstractmethod
import numpy as np
import os

from photonai_graph.GraphConversions import dense_to_networkx
from photonai_graph.util import assert_imported

try:
    from gem.embedding.static_graph_embedding import StaticGraphEmbedding
except ImportError:  # pragma: no cover
    pass


class GraphEmbeddingBase(BaseEstimator, TransformerMixin, ABC):
    _estimator_type = "transformer"
    """
    Base class for all embeddings. Implements helper functions
    used by other embeddings. Implementation based on gem
    python package.


    Parameters
    ----------
    * `embedding_dimension` [int, default=1]:
        the number of dimensions that the final embedding will have
    * `adjacency_axis` [int, default=0]:
        position of the adjacency matrix, default being zero
    * `logs` [str, default=None]:
        Path to the log data

    """

    def __init__(self,
                 embedding_dimension: int = 1,
                 adjacency_axis: int = 0,
                 logs: str = ''):
        self.embedding_dimension = embedding_dimension
        self.adjacency_axis = adjacency_axis
        self.logs = logs
        if not self.logs:
            self.logs = os.getcwd()
        assert_imported(["gem"])

    def fit(self, X, y):
        return self

    def _calculate_embedding(self, embedding: StaticGraphEmbedding, matrices: np.ndarray) -> np.ndarray:
        """Returns embedding of graphs"""
        # convert matrices
        graphs = dense_to_networkx(matrices, adjacency_axis=self.adjacency_axis)
        embedding_list = []

        for graph in graphs:
            embedding.learn_embedding(graph=graph, is_weighted=True, no_python=True)
            embedding_representation = embedding.get_embedding()
            embedding_list.append(embedding_representation)

        embedding_list = np.squeeze(np.asarray(embedding_list))

        return embedding_list

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transforms graph using Laplacian Eigenmaps Embedding"""
        embedding = self._init_embedding()
        transformed = self._calculate_embedding(embedding, X)
        transformed = np.real(transformed)
        return transformed

    @abstractmethod
    def _init_embedding(self) -> StaticGraphEmbedding:
        """Initilaize the current embedding"""
